from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

from .fitskeyword import FITSKeyword
try:
    from .feder import Feder
except ImportError:
    pass
from .patchers import *
from .astrometry import *



"""
DESCRIPTION
-----------
    For the directory provided on the command line sort the FITS files in this
    way::

        destination
            |
            |
            |
            |---'BIAS'
            |
            |---'DARK'
            |     |---exposure_time_1
            |     |---exposure_time_2, etc.
            |
            |---'FLAT'
            |      |---filter_1
            |      |    |---exposure_time_1
            |      |    |---exposure_time_2, etc.
            |      |
            |      |---filter_2, etc.
            |
            |---'LIGHT'
                    |---object_1
                    |       |---filter_1
                    |       |    |---exposure_time_1
                    |       |    |---exposure_time_2, etc.
                    |       |
                    |       |---filter_2, etc.
                    |
                    |---object_2
                    |       |---filter_1
                    |       |---filter_2, etc.
                    |
                    |---object_3, etc.
                    |
                    |---'no_object'
                            |---filter_1
                            |---filter_2, etc.

    The names in single quotes, like `'bias'`, appear exactly as written in the
    directory tree created. Names like `exposure_time_1` are replaced with a
    value, for example 30.0 if the first dark exposure time is 30.0 seconds.

    The directory ``destination/calibration/flat/R`` will contain all of the
    FITS files that are R-band flats.

    .. Note::
        This script is **NOT RECURSIVE**; it will not process files in
        subdirectories of the the directories supplied on the command line.

    .. Warning::
        Unless you explicitly supply a destination using the --destination-dir
        option the files will be copied/moved in the directory in which they
        currently exist. While this *should not* lead to information loss,
        since files are moved or copied but never deleted, you have been
        warned.

EXAMPLES
--------


"""

from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import os
import logging
import shutil

from astropy.extern.six.moves import zip as izip

from ..customlogger import console_handler, add_file_handlers
from .. import ImageFileCollection
from .. import TableTree
from . import script_helpers

UNSORTED_DIR = 'unsorted'

logger = logging.getLogger()
screen_handler = console_handler()
logger.addHandler(screen_handler)


def copy_files(files, dest, copy_or_move):
    """
    Copy a list of files to a directory

    Parameters
    ----------
    files : list of str
        List of paths of files to be copied
    dest : str
        Name of dirctory to which files should be copied
    """
    for f in files:
        copy_or_move(f, dest)


def sort_directory(directory, verbose=False,
                   destination=None,
                   no_log_destination=False,
                   script_name=None,
                   move=False):
    """
    Sort files in a directory into a tree

    Parameters
    ----------
    directory : str
        Directory whose files are to be sorted
    verbose : bool
        If True, increase logging verbosity
    destination : str, optional
        Directory into which the sorted files/directories should be placed. If
        omitted, sorting is done in the source ``directory``.
    no_log_destination : bool, optional
        Suppress logging in the destination directory. Logging cannot be
        suppressed if you are running in the destination directory.
    script_name : str, optional, default is 'sort_files'
        Name of the script calling this function; used to set the name of the
        log file in the destination.
    """
    script_name = script_name or 'sort_files'
    if destination is not None:
        working_dir = destination
        if not os.path.exists(destination):
            os.makedirs(destination)
    else:
        working_dir = directory

    if (not no_log_destination) and (destination is not None):
        add_file_handlers(logger, working_dir, script_name)

    if move:
        copy_or_move = shutil.move
    else:
        copy_or_move = shutil.copy2

    logger.info("Working on directory: %s", directory)
    logger.info("Destination directory is: %s", destination)

    default_keys = ['imagetyp', 'exptime', 'filter', 'object']
    images = ImageFileCollection(directory, keywords=default_keys)
    if not images.files:
        return
    full_table = images.summary_info
    bias = 'BIAS'
    dark = 'DARK'
    flat = 'FLAT'
    light = 'LIGHT'
    image_types = {bias: None,
                   dark: ['exptime'],
                   flat: ['filter', 'exptime'],
                   light: ['object', 'filter', 'exptime']}
    table_by_type = full_table.group_by('imagetyp')
    prepend_path = lambda path, file_list: \
                        [os.path.join(path, f) for f in file_list]
    for im_type, table in izip(table_by_type.groups.keys,
                               table_by_type.groups):
        image_type = im_type['imagetyp']
        tree_keys = image_types[image_type]
        dest_dir = os.path.join(working_dir, image_type)
        if not tree_keys:
            # bias
            source_files = prepend_path(directory, table['file'])
            this_dest = os.makedirs(dest_dir)
            copy_files(source_files, dest_dir, copy_or_move)
            continue
        mask = [False] * len(table)
        for key in tree_keys:
            mask |= table[key].mask
        if any(mask):
            source_files = prepend_path(directory, table['file'][mask])
            this_dest = os.path.join(dest_dir, UNSORTED_DIR)
            os.makedirs(this_dest)
            copy_files(source_files, this_dest, copy_or_move)
        clean_table = table[~mask]
        try:
            tree = TableTree(clean_table, tree_keys, 'file')
        except IndexError:
            continue
        for parents, children, files in tree.walk():
            if files:
                str_parents = [str(p) for p in parents]
                this_dest = os.path.join(dest_dir, *str_parents)
                os.makedirs(this_dest)
                source_files = prepend_path(directory, files)
                copy_files(source_files, this_dest, copy_or_move)


def construct_parser():
    parser = script_helpers.construct_default_parser(__doc__)
    parser.add_argument('--move', '-m', action='store_true',
                        help='Move files instead of copying them.')
    return parser


def main(arglist=None):
    """See script_helpers._main_function_docstring for actual documentation
    """
    parser = construct_parser()
    args = parser.parse_args(arglist)
    logger.debug('args are %s', vars(args))

    script_helpers.setup_logging(logger, args, screen_handler)

    add_file_handlers(logger, os.getcwd(), 'sort_files')

    # coverage misses the line below, which is actually tested..
    if not args.dir:    # pragma: no cover
        parser.error('No directory specified')

    do_not_log_in_destination = \
        script_helpers.handle_destination_dir_logging_check(args)

    sort_directory(args.dir[0],
                   verbose=args.verbose,
                   destination=args.destination_dir,
                   no_log_destination=do_not_log_in_destination,
                   move=args.move)

main.__doc__ = script_helpers._main_function_docstring(__name__)



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

from os import path
from datetime import datetime
import logging
from socket import timeout

import numpy as np
import astropy.io.fits as fits
from astropy.time import Time
from astropy.coordinates import Angle, name_resolve, SkyCoord, AltAz
from astropy import units as u
from astropy.table import Table

from astropy.extern.six.moves.urllib import parse as urlparse

try:
    from .feder import Feder
    feder = Feder()
except ImportError:
    feder = None
    pass

from ..image_collection import ImageFileCollection
from .fitskeyword import FITSKeyword

logger = logging.getLogger(__name__)


#__all__ = ['patch_headers', 'add_object_info', 'add_ra_dec_from_object_name']


def IRAF_image_type(image_type):
    """
    Convert MaximDL default image type names to IRAF

    Parameters
    ----------
    image_type : str
        Value of the FITS header keyword IMAGETYP; acceptable values are
        below in Notes.

    Returns
    -------
    str
        IRAF image type (one of 'BIAS', 'DARK', 'FLAT' or 'LIGHT')

    Notes
    -----
    The MaximDL default is, e.g. 'Bias Frame', which IRAF calls
    'BIAS'. Can safely be called with an IRAF-style image_type.
    """
    return image_type.split()[0].upper()


def _lst_from_obstime(obstime):
    try:
        LST = obstime.sidereal_time('apparent',
                                    longitude=feder.site.longitude)
    except IndexError:
        # We are outside the range of the IERS table installed with astropy,
        # so get a newer one.
        from astropy.utils import iers
        from astropy.utils.data import download_file
        iers_a = iers.IERS_A.open(download_file(iers.IERS_A_URL,
                                  cache=True))
        obstime.delta_ut1_utc = obstime.get_delta_ut1_utc(iers_a)
        LST = obstime.sidereal_time('apparent',
                                    longitude=feder.site.longitude)
    return LST


def add_time_info(header, history=False):
    """
    Add JD, MJD, LST to FITS header

    Parameters
    ----------
    header : astropy..io.fits.Header
        FITS header to be modified.
    history : bool
        If `True`, write history for each keyword changed.
    """
    dateobs = Time(header['date-obs'], scale='utc')
    feder.JD_OBS.value = dateobs.jd
    feder.MJD_OBS.value = dateobs.mjd

    LST_tmp = _lst_from_obstime(dateobs)

    feder.LST.value = LST_tmp.to_string(unit=u.hour, sep=':', precision=4,
                                        pad=True)

    for keyword in feder.keywords_for_all_files:
        keyword.add_to_header(header, history=history)
        logger.info(keyword.history_comment())


def add_object_pos_airmass(header, history=False):
    """
    Add object information, such as RA/Dec and airmass.

    Parameters
    ----------
    header : astropy..io.fits.Header
        FITS header to be modified.
    history : bool
        If `True`, write history for each keyword changed.

    Notes
    -----
    Has side effect of setting feder site JD to JD-OBS, which means it
    also assume JD.value has been set.

    """
    # not sure why coverage is not picking up both branches, but it is not, so
    # marking it no cover
    if feder.JD_OBS.value is None:   # pragma: no cover
        raise ValueError('Need to set JD_OBS.value '
                         'before calling.')

    try:
        feder.RA.set_value_from_header(header)
    except ValueError:
        raise ValueError("No RA is present.")

    feder.DEC.set_value_from_header(header)
    feder.RA.value = feder.RA.value.replace(' ', ':')
    feder.DEC.value = feder.DEC.value.replace(' ', ':')

    obj_coord2 = SkyCoord(feder.RA.value, feder.DEC.value,
                          unit=(u.hour, u.degree), frame='fk5')

    obstime = Time(feder.MJD_OBS.value, format='mjd')
    alt_az = obj_coord2.transform_to(AltAz(obstime=obstime,
                                           location=feder.site))

    feder.ALT_OBJ.value = round(alt_az.alt.degree, 5)
    feder.AZ_OBJ.value = round(alt_az.az.degree, 5)
    feder.AIRMASS.value = round(1 / np.cos(np.pi / 2 - alt_az.alt.radian), 3)

    # TODO: replace the LST calculation
    LST = _lst_from_obstime(obstime)
    HA = LST.hour - obj_coord2.ra.hour
    HA = Angle(HA, unit=u.hour)

    feder.HA.value = HA.to_string(unit=u.hour, sep=':')

    for keyword in feder.keywords_for_light_files:
        if keyword.value is not None:
            keyword.add_to_header(header, history=history)
            logger.info(keyword.history_comment())


def get_software_name(header, file_name=None, use_observatory=None):
    """
    Determine the name of the software that created FITIS header

    Parameters
    ----------

    header : astropy.io.fits Header
        Header from a FITS extension/hdu

    file_name : str, optional
        Name of the file containing this header; used to add information to
        error/warning messages.

    use_observatory : msumastro.Feder instance, optional
        Object that contains names of FITS keywords that might be present and
        contain name of the software that made this header. The default value
        is the instance defined at the beginning of this module

    Returns
    -------

    msumastro.feder.Software object
    """
    fits_file = file_name or ''
    observatory = use_observatory or feder

    known_software_keywords = observatory.software_FITS_keywords

    software_name_in_header = FITSKeyword(name=known_software_keywords[0],
                                          synonyms=known_software_keywords)
    software_name_in_header.set_value_from_header(header)

    try:
        software = observatory.software[software_name_in_header.value]
    except KeyError:
        raise KeyError('Software named {0} not recognized in header '
                       'for file {1}'.format(software_name_in_header.value,
                                             fits_file))
    return software


def purge_bad_keywords(header, history=False, force=False, file_name=None):
    """
    Remove keywords from FITS header that may be incorrect

    Parameters
    ----------
    header : astropy.io.fits.Header
        Header from which the bad keywords (as defined by the software that
        recorded the image) should be purged.

    history : bool
        If `True` write detailed history for each keyword removed.

    force : bool
        If `True`, force keywords to be purged even if the FITS header
        indicates it has already been purged.

    file_name : str, optional
        Name of file containing the header; if provided it is used to generate
        more informative log messages.
    """
    software = get_software_name(header, file_name=file_name)

    try:
        purged = header[software.purged_flag_keyword]
    except KeyError:
        purged = False

    if purged and not force:
        warn_msg = 'Not removing headers from {0} again, '
        warn_msg += 'set force=True to force removal.'
        logger.warn(warn_msg.format(file_name))
        return

    for keyword in software.bad_keywords:
        try:
            comment = ('Deleted keyword ' + keyword +
                       ' with value ' + str(header[keyword]))
        except KeyError:
            continue
        logger.info(comment)
        del header[keyword]
        if history:
            header.add_history(comment)

    header[software.purged_flag_keyword] = (True,
                                            'Have bad keywords been removed?')


def change_imagetype_to_IRAF(header, history=True):
    """
    Change IMAGETYP to default used by IRAF

    Parameters
    ----------
    header : astropy.io.fits.Header
        Header object in which image type is to be changed.

    history : bool, optional
        If `True`, add history of keyword modification to `header`.
    """
    imagetype = 'imagetyp'  # FITS keyword name is truncated at 8 chars
    current_type = header[imagetype]
    IRAF_type = IRAF_image_type(current_type)
    if current_type != IRAF_type:
        header[imagetype] = IRAF_type
        comment = 'Changed {0} from {1} to {2}'.format(imagetype.upper(),
                                                       current_type,
                                                       IRAF_type)
        logger.info(comment)
        if history:
            header.add_history(comment)


def add_image_unit(header, history=True):
    """
    Add unit of image to header.

    Parameters
    ----------
    header : astropy.io.fits.Header
        Header object in which image type is to be changed.

    history : bool, optional
        If `True`, add history of keyword modification to `header`.
    """
    instrument_key = 'instrume'
    instrument = feder.instruments[header[instrument_key]]
    if instrument.image_unit is not None:
        unit_string = instrument.image_unit.to_string()
        comment = 'Set image data unit to {}'.format(unit_string)
        header['BUNIT'] = unit_string
        print('UNIT STRING IS =========> {}'.format(unit_string))
        logger.info(comment)
        if history:
            header.add_history(comment)


def list_name_is_url(name):
    may_be_url = urlparse.urlparse(name)
    return (may_be_url.scheme and may_be_url.netloc)


def read_object_list(directory=None, input_list=None,
                     skip_consistency_check=False, check_radius=20.0,
                     skip_lookup_from_object_name=False):
    """
    Read a list of objects from a text file.

    Parameters
    ----------
    directory : str
        Directory containing the file. Default is the current directory, ``.``

    input_list : str, optional
        Name of the file or URL of file. Default value is ``obsinfo.txt``. If
        the name is a URL the directory argument is ignored.

    skip_consistency_check : bool optional
        If ``True``, skip checking whether objects on the list have unique
        coordinates given `check_radius`.

    check_radius : float, optional
        Match radius, in arcminutes. Objects on the list must be separated by
        an angular distance greater than this for the list to be
        self-consistent.

    skip_lookup_from_object_name : bool, optional
        Set to ``True`` to skip lookup of coordinates from Simbad if RA/Dec
        are not in the object file.

    Notes
    -----

    There are two file formats; one contains just a list of objects, the
    other has an RA and Dec for each object.

    In both types any lines that start with ``#`` are ignored and treated
    as comments.

    **File with list of objects only:**
        + Object coordinates are determined by lookup with
          `Simbad <http://simbad.u-strasbg.fr/simbad/>`_. You should make sure
          the object names you use are known to simbad.
        + The first non-comment line MUST be the word ``object`` and only
          the word ``object``. It is case sensitive; ``Object`` or ``OBJECT``
          will not work.
        + Remaining line(s) are name(s) of object(s), one per line. Case does
          **not** matter for object name.
        + Example::

            # my list is below
            object
            m101
            sz lyn
            # the next object is after this comment
            RR LYR

    **File with object name and position:**
        + RA and Dec **must be J2000**.
        + RA **must be given in hours**, though it can be either sexagesimal
          (e.g. ``19:25:27.9``) or decimal (e.g. ``19.423861``).
        + Dec **must be given in degrees**, though it can be either sexagesimal
          (e.g. ``42:47:3.69``) or decimal (e.g. ``42.7843583``)
        + The first non-comment line MUST be these words: ``object,RA,Dec``.
          These are column headings for your file. It is **not** case
          sensitive; for example, using ``DEC`` instead of ``Dec`` will work.
        + Each remaining line should be an object name, object RA and Dec.
          Case does **not** matter for object name.
        + Example::

            # my list with RA and Dec
            # RA and Dec are assumed to be J2000
            # RA MUST BE IN HOURS
            # DEC MUST BE IN DEGREES
            object,RA,Dec
            m101,14:03:12.583, +54:20:55.50
            # note that the leading sign for the Dec is optional if Dec is
            # positive
            sz lyn,08:09:35.748, 44:28:17.61
            # You can mix sexagesimal and decimal RA/Dec.
            RR Lyr, 19.423861,42.7843583

    """
    def normalize_column_name(key, table):
        """
        Find any column in table whose name matches, aside from case, key
        and change the name of the column to key
        """
        contains_col = lambda key, names: key.lower() in [name.lower() for name
                                                          in names]

        if not contains_col(key, table.columns):
            raise KeyError('Keyword {0} not found in table'.format(key))

        for column in table.columns:
            if ((key.lower() == column.lower()) and (key != column)):
                table.rename_column(column, key)
                break

    if directory is None:
        directory = '.'
    list_name = input_list if input_list is not None else 'obsinfo.txt'

    if not list_name_is_url(list_name):
        full_name = path.join(directory, list_name)
    else:
        full_name = list_name

    objects = Table.read(full_name,
                         format='ascii',
                         comment='#',
                         delimiter=',')

    try:
        normalize_column_name('object', objects)
    except KeyError as e:
        logger.debug('%s', e)
        raise RuntimeError(
              'No column named object found in file {}'.format(list_name))

    try:
        normalize_column_name('RA', objects)
        normalize_column_name('Dec', objects)
        RA = objects['RA']
        Dec = objects['Dec']
    except KeyError:
        RA = None
        Dec = None
        ra_dec = None

    object_names = objects['object']

    if (RA is not None) and (Dec is not None):
        default_angle_units = (u.hour, u.degree)
        ra_dec = SkyCoord(RA, Dec, unit=default_angle_units, frame='fk5')
    else:
        if skip_lookup_from_object_name:
            ra_dec = None
        else:
            try:
                ra_dec = [SkyCoord.from_name(obj) for obj in object_names]
            except (name_resolve.NameResolveError, timeout) as e:
                logger.error('Unable to do lookup of object positions')
                logger.error(e)
                raise name_resolve.NameResolveError('Unable to do lookup of '
                                                    'object positions')
            ra = []
            dec = []
            for a_coord in ra_dec:
                ra.append(a_coord.ra.radian)
                dec.append(a_coord.dec.radian)
            ra_dec = SkyCoord(ra, dec, unit=(u.radian, u.radian), frame='fk5')

    if skip_consistency_check and skip_lookup_from_object_name:
        return object_names, ra_dec

    if skip_consistency_check or not ra_dec:
        return object_names, ra_dec

    # sanity check object list--the objects should not be so close together
    # that any pair is within match radius of each other.
    logger.debug('Testing object list for self-consistency')

    # need 2nd neighbor below or objects will match themselves
    try:
        matches, d2d, d3d = ra_dec.match_to_catalog_sky(ra_dec,
                                                        nthneighbor=2)
        bad_object_list = (d2d.arcmin < check_radius).any()
    except IndexError:
        # There was only one item in the table...so no object can
        # be duplicated, so make a fake distance that doesn't match
        bad_object_list = False

    if bad_object_list:
        err_msg = ('Object list {} in directory {} contains at least one '
                   'pair of objects that '
                   'are closer together than '
                   'match radius {} arcmin').format(input_list,
                                                    directory,
                                                    check_radius)
        logger.error(err_msg)
        raise RuntimeError(err_msg)
    logger.debug('Passed: object list is self-consistent')

    return object_names, ra_dec


def history(function, mode=None, time=None):
    """
    Construct nicely formatted start/end markers in FITS history.

    Parameters
    ----------
    function : func
        Function calling `history`
    mode : str, 'begin' or 'end'
        A different string is produced for the beginning and the end. Default
        is 'begin'.
    time : datetime
        If not set, defaults to current date/time.
    """
    mode = mode or 'begin'
    if mode == 'begin':
        marker = '+'
    elif mode == 'end':
        marker = '-'
    else:
        raise ValueError('mode must be "begin" or "end"')

    if time is None:
        time = datetime.now()

    marker *= 5
    return "%s %s %s history on %s %s" % (marker, mode.upper(),
                                          function.__name__, time,
                                          marker)


def patch_headers(dir=None,
                  new_file_ext=None,
                  save_location=None,
                  overwrite=False,
                  purge_bad=True,
                  add_time=True,
                  add_apparent_pos=True,
                  add_overscan=True,
                  fix_imagetype=True,
                  add_unit=True):
    """
    Add minimal information to Feder FITS headers.

    Parameters
    ----------
    dir : str, optional
        Directory containing the files to be patched. Default is the current
        directory, ``.``

    new_file_ext : str, optional
        Name added to the FITS files with updated header information. It is
        added to the base name of the input file, between the old file name
        and the `.fit` or `.fits` extension. Default is 'new'.

    save_location : str, optional
        Directory to which the patched files should be written, if not `dir`.

    overwrite : bool, optional
        Set to `True` to replace the original files.

    purge_bad : bool, optional
        Remove "bad" keywords form header before any other processing. See
        :func:`purge_bad_keywords` for details.

    add_time : bool, optional
        If ``True``, add time information (e.g. JD, LST); see
        :func:`add_time_info` for details.

    add_apparent_pos : bool, optional
        If ``True``, add apparent position (e.g. alt/az) to headers. See
        :func:`add_object_pos_airmass` for details.

    add_overscan : bool, optional
        If ``True``, add overscan keywords to the headers. See
        :func:`add_overscan_header` for details.

    fix_imagetype : bool, optional
        If ``True``, change image types to IRAF-style. See
        :func:`change_imagetype_to_IRAF` for details.

    add_unit : bool, optional
        If ``True``, add image unit to FITS header.
    """
    dir = dir or '.'
    if new_file_ext is None:
        new_file_ext = 'new'

    images = ImageFileCollection(location=dir, keywords=['imagetyp'])

    for header, fname in images.headers(save_with_name=new_file_ext,
                                        save_location=save_location,
                                        overwrite=overwrite,
                                        do_not_scale_image_data=True,
                                        return_fname=True):
        run_time = datetime.now()

        logger.info('START PATCHING FILE: {0}'.format(fname))

        header.add_history(history(patch_headers, mode='begin',
                                   time=run_time))
        header.add_history('patch_headers.py modified this file on %s'
                           % run_time)
        try:
            # each of the next 3 lines checks for presences of something
            get_software_name(header)  # is there some software?
            header['instrume']  # is there an instrument?
            header['imagetyp']  # is there an image type?

            if purge_bad:
                purge_bad_keywords(header, history=True, file_name=fname)

            if fix_imagetype:
                change_imagetype_to_IRAF(header, history=True)

            if add_time:
                add_time_info(header, history=True)

            if add_overscan:
                add_overscan_header(header, history=True)

            if add_unit:
                add_image_unit(header, history=True)

            # add_apparent_pos_airmass can raise a ValueError, do it last.
            if add_apparent_pos and (header['imagetyp'] == 'LIGHT'):
                add_object_pos_airmass(header,
                                       history=True)

        except (KeyError, ValueError) as e:
            warning_msg = ('********* FILE NOT PATCHED *********'
                           'Stopped patching header of {0} because of '
                           '{1}: {2}'.format(fname, type(e).__name__, e))
            logger.warn(warning_msg)
            header.add_history(warning_msg)
            continue
        finally:
            header.add_history(history(patch_headers, mode='end',
                               time=run_time))
            logger.info('END PATCHING FILE: {0}'.format(fname))


def add_overscan_header(header, history=True):
    """
    Add overscan information to a FITS header.

    Parameters
    ----------
    header : astropy.io.fits.Header
        Header object to which overscan is to be added.

    history : bool, optional
        If `True`, add history of keyword modification to `header`.

    Returns
    -------
    list of str
        List of the keywords added to the header by this function.
    """
    image_dim = [header['naxis1'], header['naxis2']]
    instrument = feder.instruments[header['instrume']]
    overscan_present = instrument.has_overscan(image_dim)
    modified_keywords = []
    if overscan_present:
        overscan_region = feder.BIASSEC
        trim_region = feder.TRIMSEC
        overscan_region.value = instrument.useful_overscan
        trim_region.value = instrument.trim_region
        overscan_region.add_to_header(header, history=history)
        trim_region.add_to_header(header, history=history)
        modified_keywords.extend([overscan_region, trim_region])

    for keyword in modified_keywords:
        logger.info(keyword.history_comment())

    return modified_keywords


def add_object_info(directory=None,
                    object_list=None,
                    object_list_dir=None,
                    match_radius=20.0, new_file_ext=None,
                    save_location=None,
                    overwrite=False, detailed_history=True):
    """
    Add object information to FITS files that contain pointing information
    given a list of objects.

    Parameters
    ----------
    directory : str
        Directory containing the FITS files to be fixed. Default is the
        current directory, ``.``.

    object_list : str, optional
        Name of file containing list of objects. Default is set by
        :func:`read_object_list` which also explains the format of this file.

    object_list_dir : str, optional
        Directory in which the `object_list` is contained. Default is
        `directory`.

    match_radius : float, optional
        Maximum distance, in arcmin, between the RA/Dec of the image and a
        particular object for the image to be considered an image of that
        object.

    new_file_ext : str, optional
        Name added to the FITS files with updated header information. It is
        added to the base name of the input file, between the old file name
        and the `.fit` or `.fits` extension. Default is 'new'.

    save_location : str, optional
        Directory to which the patched files should be written, if not `dir`.

    overwrite : bool, optional
        Set to `True` to replace the original files.
    """
    directory = directory or '.'
    if new_file_ext is None:
        new_file_ext = 'new'

    images = ImageFileCollection(directory,
                                 keywords=['imagetyp', 'ra',
                                           'dec', 'object'])
    im_table = images.summary_info

    object_dir = directory if object_list_dir is None else object_list_dir

    logger.debug('About to read object list')
    try:
        object_names, ra_dec = read_object_list(object_dir,
                                                input_list=object_list)
    except IOError:
        warn_msg = 'No object list in directory {0}, skipping.'
        logger.warn(warn_msg.format(directory))
        return
    except name_resolve.NameResolveError:
        logger.error('Unable to add objects--name resolve error')
        return

    object_names = np.array(object_names)

    # I want rows which...
    #
    # ...have no OBJECT...
    needs_object = im_table['object'].mask
    # ...and have coordinates.
    needs_object &= ~ (im_table['ra'].mask | im_table['dec'].mask)

    logger.debug('Looking for objects for %s images', needs_object.sum())
    # Qualifying rows need a search for a match.
    # the search returns a match for every row provided, but some matches
    # may be farther away than desired, so...
    #
    # ...`and` the previous index mask with those that matched, and
    # ...construct list of object names for those images.
    default_angle_units = (u.hour, u.degree)

    img_pos = SkyCoord(im_table['ra'][needs_object],
                       im_table['dec'][needs_object],
                       unit=default_angle_units,
                       frame='fk5')
    match_idx, d2d, d3d = img_pos.match_to_catalog_sky(ra_dec)
    good_match = (d2d.arcmin <= match_radius)
    found_object = np.array(needs_object)
    found_object[needs_object] = good_match
    matched_object_name = object_names[match_idx][good_match]

    no_match_found = needs_object & ~found_object
    if no_match_found.any():
        for fname in np.array(images.files)[no_match_found]:
            warn_msg = "No object found for image {0}".format(fname)
            logger.warn(warn_msg)

    if not found_object.any():
        logger.info('NO OBJECTS MATCHED TO IMAGES IN: {0}'.format(directory))
        return

    im_table['file'].mask = ~found_object

    for idx, (header, fname) in enumerate(images.headers(save_with_name=new_file_ext,
                                          overwrite=overwrite,
                                          save_location=save_location,
                                          return_fname=True)):

        logger.info('START ATTEMPTING TO ADD OBJECT to: {0}'.format(fname))

        object_name = matched_object_name[idx]
        logger.debug('Found matching object named %s', object_name)
        obj_keyword = FITSKeyword('object', value=object_name)
        obj_keyword.add_to_header(header, history=True)
        logger.info(obj_keyword.history_comment())
        logger.info('END ATTEMPTING TO ADD OBJECT to: {0}'.format(fname))


def add_ra_dec_from_object_name(directory=None, new_file_ext=None,
                                object_list=None, object_list_dir=None):
    """
    Add RA/Dec to FITS file that has object name but no pointing.

    Parameters
    ----------
    dir : str, optional
        Directory containing the files to be patched. Default is the current
        directory, ``.``

    new_file_ext : str, optional
        Name added to the FITS files with updated header information. It is
        added to the base name of the input file, between the old file name
        and the `.fit` or `.fits` extension. Default is 'new'.

    object_list : str, optional
        Name of file containing list of objects. Default is set by
        :func:`read_object_list` which also explains the format of this file.

    object_list_dir : str, optional
        Directory in which the `object_list` is contained. Default is
        `directory`.

    """
    directory = directory or '.'
    if new_file_ext is None:
        new_file_ext = 'new'
    images = ImageFileCollection(directory,
                                 keywords=['imagetyp', 'ra',
                                           'dec', 'object'])
    summary = images.summary_info
    missing_dec = summary[(np.logical_not(summary['object'].mask)) &
                          (summary['ra'].mask) &
                          (summary['dec'].mask) &
                          (summary['object'] != '') &
                          (summary['imagetyp'] == 'LIGHT')]

    if not missing_dec:
        return

    objects = np.unique(missing_dec['object'])

    try:
        object_list, ra_dec_list = read_object_list(object_list_dir,
                                                    input_list=object_list)
    except IOError:
        object_list = []
        ra_dec_list = []
        object_dict = {}

    if len(object_list) and len(ra_dec_list):
        object_dict = {obj: ra_dec for obj, ra_dec in zip(object_list,
                                                          ra_dec_list)}

    # checks prior to this mean this loop always happens at least once,
    # which confuses coverage
    for object_name in objects:  # pragma: nobranch
        try:
            object_coords = object_dict[object_name]
        except KeyError:
            try:
                object_coords = SkyCoord.from_name(object_name)
            except (name_resolve.NameResolveError, timeout) as e:
                logger.warning('Unable to lookup position for %s', object_name)
                logger.warning(e)
                return

        common_format_keywords = {'sep': ':',
                                  'precision': 2,
                                  'pad': True}
        feder.RA.value = object_coords.ra.to_string(unit=u.hour,
                                                    **common_format_keywords)
        feder.DEC.value = object_coords.dec.to_string(unit=u.degree,
                                                      alwayssign=True,
                                                      **common_format_keywords)
        these_files = missing_dec[missing_dec['object'] == object_name]
        for image in these_files:
            full_name = path.join(directory, image['file'])
            hdulist = fits.open(full_name, do_not_scale_image_data=True)
            header = hdulist[0].header
            feder.RA.add_to_header(header, history=True)
            feder.DEC.add_to_header(header, history=True)
            if new_file_ext:
                base, ext = path.splitext(full_name)
                new_file_name = base + new_file_ext + ext
                overwrite = False
            else:
                new_file_name = full_name
                overwrite = True
            hdulist.writeto(new_file_name, clobber=overwrite)
            hdulist.close()



"""
DESCRIPTION
-----------
    For each directory provided on the command line add
    astrometry to the light files (those with ``IMAGETYP='LIGHT'`` in
    the FITS header).

    By default, astrometry is added only for those files with pointing
    information in the header (specifically, RA and Dec) because blind
    astrometry is fairly slow. It may be faster to insert RA/Dec into
    those files before doing astrometry.

    The functions called by this script set the WCS reference pixel
    to the center of the image, which turns out to make aligning images
    a little easier.

    For more control over the parameters see :func:`add_astrometry`
    and for even more control, :func:`call_astrometry`.

    .. Note::
        This script is **NOT RECURSIVE**; it will not process files in
        subdirectories of the the directories supplied on the command line.

    .. WARNING::
        This script OVERWRITES the image files in the directories
        specified on the command line unless you use the --destination-dir
        option.

EXAMPLES
--------

    Invoking this script from the command line::

        run_astrometry.py /my/folder/of/images

    To work on the same folder from within python, do this::

        from msumastro.scripts import run_astrometry
        run_astrometry.main(['/my/folder/of/images'])
"""
from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import shutil
from os import path, getcwd
import logging

import numpy as np

from astropy.io import fits
import astropy.units as u
from astropy.coordinates import SkyCoord

from ..customlogger import console_handler, add_file_handlers
from ..header_processing import astrometry as ast
from .. import ImageFileCollection
from ..image import ImageWithWCS
from .script_helpers import (construct_default_parser, setup_logging,
                             handle_destination_dir_logging_check,
                             _main_function_docstring)

logger = logging.getLogger()
screen_handler = console_handler()
logger.addHandler(screen_handler)


def astrometry_for_directory(directories,
                             destination=None,
                             no_log_destination=False,
                             blind=False):
    """
    Add astrometry to files in list of directories

    Parameters
    ----------

    directories : str or list of str
        Directory or directories whose FITS files are to be processed.

    blind : bool, optional
        Set to True to force blind astrometry. False by default because
        blind astrometry is slow.
    """

    for currentDir in directories:
        images = ImageFileCollection(currentDir,
                                     keywords=['imagetyp', 'object',
                                               'wcsaxes', 'ra', 'dec'])
        summary = images.summary_info
        if len(summary) == 0:
            continue
        logger.debug('\n %s', '\n'.join(summary.pformat()))
        lights = summary[((summary['imagetyp'] == 'LIGHT') &
                          (summary['wcsaxes'].mask))]

        working_dir = destination if destination is not None else currentDir
        if (not no_log_destination) and (destination is not None):
            add_file_handlers(logger, working_dir, 'run_astrometry')

        logger.debug('About to loop over %d files', len(lights['file']))
        for light_file in lights['file']:
            if ((destination is not None) and (destination != currentDir)):
                src = path.join(currentDir, light_file)
                shutil.copy(src, destination)

            original_fname = path.join(working_dir, light_file)
            img = ImageWithWCS(original_fname)
            try:
                ra = img.header['ra']
                dec = img.header['dec']
                ra_dec = (ra, dec)
            except KeyError:
                ra_dec = None

            if (ra_dec is None) and (not blind):
                root, ext = path.splitext(original_fname)
                f = open(root + '.blind', 'wb')
                f.close()
                continue

            astrometry = ast.add_astrometry(img.fitsfile.filename(),
                                            ra_dec=ra_dec,
                                            note_failure=True,
                                            overwrite=True)

            with fits.open(img.fitsfile.filename(),
                           do_not_scale_image_data=True) as f:
                try:
                    del f[0].header['imageh'], f[0].header['imagew']
                    f.writeto(img.fitsfile.filename(), clobber=True)
                except KeyError:
                    pass

            if astrometry and ra_dec is None:
                root, ext = path.splitext(original_fname)
                img_new = ImageWithWCS(original_fname)
                ra_dec = img_new.wcs_pix2world(np.trunc(np.array(img_new.shape)
                                               / 2))
                # RA/Dec are in degrees. Convert them to sexagesimal for
                # output. Yuck, but makes it easier for existing code to
                # handle.
                # Note that FK5 is J2000.
                coords = SkyCoord(*ra_dec, unit=(u.degree, u.degree),
                                  frame='fk5')
                img_new.header['RA'] = coords.ra.to_string(unit=u.hour, sep=':')
                img_new.header['DEC'] = coords.dec.to_string(sep=':')
                img_new.save(img_new.fitsfile.filename(), overwrite=True)


def construct_parser():
    parser = construct_default_parser(__doc__)

    blind_help = 'Turn ON blind astrometry; '
    blind_help += 'disabled by default because it is so slow.'
    parser.add_argument('-b', '--blind',
                        help=blind_help, action='store_true')

    return parser


def main(arglist=None):
    """See script_helpers._main_function_docstring for actual documentation
    """

    parser = construct_parser()
    args = parser.parse_args(arglist)

    setup_logging(logger, args, screen_handler)

    add_file_handlers(logger, getcwd(), 'run_astrometry')

    do_not_log_in_destination = handle_destination_dir_logging_check(args)

    astrometry_for_directory(args.dir,
                             destination=args.destination_dir,
                             blind=args.blind,
                             no_log_destination=do_not_log_in_destination)

main.__doc__ = _main_function_docstring(__name__)



"""
DESCRIPTION
-----------
    For each directory provided on the command line create a table
    in that directory with one row for each FITS file in the directory.
    The columns are FITS keywords extracted from the header of each
    file.

    The list of default keywords extracted is available through the command
    line option ``--list-default``.

    .. Note::
        This feature is available only from the command line.

    For more control over the parameters see :func:`triage_fits_files`

    .. Note::
        This script is **NOT RECURSIVE**; it will not process files in
        subdirectories of the the directories supplied on the command line.


EXAMPLES
--------

    Invoking this script from the command line::

        python run_triage.py /my/folder/of/images

    Get list of default keywords included in summary table::

        python run_triage.py --list-default

    To work on the same folder from within python, do this::

        from msumastro.scripts import run_triage
        run_triage.main(['/my/folder/of/images'])
        # or...
        run_triage.main(['--list-default'])

"""

from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import os
from argparse import ArgumentParser
from sys import exit
import logging

from astropy.table import Table, Column
import numpy as np

from ..customlogger import console_handler, add_file_handlers
from ..header_processing.feder import Feder
from .. import ImageFileCollection
from . import script_helpers

logger = logging.getLogger()
screen_handler = console_handler()
logger.addHandler(screen_handler)


class DefaultFileNames(object):
    def __init__(self):
        self.object_file_name = 'NEEDS_OBJECT_NAME.txt'
        self.pointing_file_name = 'NEEDS_POINTING_INFO.txt'
        self.filter_file_name = 'NEEDS_FILTER.txt'
        self.output_table = 'Manifest.txt'
        self.astrometry_file_name = 'NEEDS_ASTROMETRY.txt'

    def as_dict(self):
        return self.__dict__


def write_list(dir, file, info, column_name=None):
    col_name = column_name or 'File'
    temp_table = Table(data=[info],
                       names=[col_name])
    temp_table.write(os.path.join(dir, file),
                     format='ascii')


def contains_maximdl_imagetype(image_collection):
    """
    Check an image file collection for MaxImDL-style image types
    """
    import re
    file_info = image_collection.summary_info

    if file_info['imagetyp'].mask.any():
        logger.warn('One or more image is missing IMAGETYP in header')

    image_types = ' '.join([typ for typ in file_info['imagetyp'].compressed()])

    if re.search('[fF]rame', image_types) is not None:
        return True
    else:
        return False


def get_column_name_case_insensitive(name, column_names):
    """
    Return the column name that matches name in a case-insensitive comparison.

    Parameters
    ----------

    name : str
        Name for which match is desired.

    column_names : list
        List of column names.

    Returns
    -------

    str or None
        name of first matching column or ``None``, if no match is found.
    """
    for cname in column_names:
        if name.lower() == cname.lower():
            return cname
    else:
        return ''


def triage_fits_files(dir=None, file_info_to_keep=None):
    """
    Check FITS files in a directory for deficient headers

    `dir` is the name of the directory to search for files.

    `file_info_to_keep` is a list of the FITS keywords to get values
    for for each FITS file in `dir`.
    """
    dir = dir or '.'
    all_file_info = file_info_to_keep or ['imagetyp', 'object',
                                          'filter', 'wcsaxes']
    feder = Feder()
    RA = feder.RA
    if ((not (set(RA.names) <= set(all_file_info))) and
       (all_file_info != '*')):
        all_file_info.extend(RA.names)

    images = ImageFileCollection(dir, keywords=all_file_info)
    file_info = images.summary_info

    # check for bad image type and halt until that is fixed.
    if contains_maximdl_imagetype(images):
        raise ValueError(
            'Correct MaxImDL-style image types before proceeding.')

    file_needs_filter = \
        list(images.files_filtered(imagetyp='light',
                                   filter=None))
    file_needs_filter += \
        list(images.files_filtered(imagetyp='flat',
                                   filter=None))

    file_needs_object_name = \
        list(images.files_filtered(imagetyp='light',
                                   object=None))

    lights = file_info[file_info['imagetyp'] == 'LIGHT']
    file_needs_pointing = []
    file_needs_astrometry = []
    if lights:
        has_no_ra = np.array([True] * len(lights))
        has_no_ha = np.array([True] * len(lights))
        for ra_name in RA.names:
            col_name = \
                get_column_name_case_insensitive(ra_name, lights.colnames)
            try:
                has_no_ra &= (lights[col_name].mask)
            except KeyError:
                pass
        for ha_name in feder.HA.names:
            col_name = \
                get_column_name_case_insensitive(ha_name, lights.colnames)
            try:
                has_no_ha &= lights[col_name].mask
            except KeyError:
                pass

        file_needs_astrometry = list(lights['file'][lights['wcsaxes'].mask])
        needs_minimal_pointing = has_no_ha | has_no_ra
        file_needs_pointing = list(lights['file'][needs_minimal_pointing])

    full_path = os.path.abspath(dir)
    path_column = Column(data=[full_path] * len(file_info), name='Source path')
    containing_dir = os.path.basename(full_path)
    containing_dir_col = Column(data=[containing_dir] * len(file_info),
                                name='Source directory')
    file_info.add_columns([path_column, containing_dir_col])

    dir_info = {'files': file_info,
                'needs_filter': file_needs_filter,
                'needs_pointing': file_needs_pointing,
                'needs_object_name': file_needs_object_name,
                'needs_astrometry': file_needs_astrometry}

    return dir_info


def triage_directories(directories,
                       keywords=None,
                       all_keywords=False,
                       object_file_name=None,
                       pointing_file_name=None,
                       filter_file_name=None,
                       astrometry_file_name=None,
                       output_table=None,
                       destination=None,
                       no_log_destination=False):

    for currentDir in directories:
        if destination is not None:
            target_dir = destination
        else:
            target_dir = currentDir
        if (not no_log_destination) and (destination is not None):
            add_file_handlers(logger, destination, 'run_triage')
        logger.info('Examining directory %s', currentDir)
        if keywords:
            # force a copy...
            use_keys = list(keywords)
        else:
            use_keys = []

        if all_keywords:
            try:
                use_keys += ['*']
            except TypeError:
                use_keys = '*'

        result = triage_fits_files(currentDir, file_info_to_keep=use_keys)
        outfiles = [pointing_file_name, filter_file_name,
                    object_file_name, output_table]
        for fil in [outfile for outfile in outfiles if outfile is not None]:
            try:
                os.remove(os.path.join(currentDir, fil))
            except OSError:
                pass

        need_pointing = result['needs_pointing']
        need_filter = result['needs_filter']
        need_object_name = result['needs_object_name']
        need_astrometry = result['needs_astrometry']

        if need_pointing and pointing_file_name is not None:
            write_list(target_dir, pointing_file_name, need_pointing)
        if need_filter and filter_file_name is not None:
            write_list(target_dir, filter_file_name, need_filter)
        if need_object_name and object_file_name is not None:
            write_list(target_dir, object_file_name,
                       need_object_name)
        if need_astrometry and astrometry_file_name is not None:
            write_list(target_dir, astrometry_file_name, need_astrometry)

        tbl = result['files']
        if ((len(tbl) > 0) and (output_table is not None)):
            tbl.write(os.path.join(target_dir, output_table),
                      format='ascii', delimiter=',')


def construct_parser():

    parser = ArgumentParser()
    script_helpers.setup_parser_help(parser, __doc__)
    # allow for no directories below so that -l option
    # can be used without needing to specify a directory
    script_helpers.add_directories(parser, '*')
    script_helpers.add_debug(parser)
    script_helpers.add_verbose(parser)
    script_helpers.add_destination_directory(parser)
    script_helpers.add_no_log_destination(parser)
    script_helpers.add_console_output_args(parser)

    key_help = 'FITS keyword to add to table in addition to the defaults; '
    key_help += 'for multiple keywords use this option multiple times.'
    parser.add_argument('-k', '--key', action='append',
                        help=key_help, default=[])

#    no_default_help = ('Do not include default list of keywords in table'
#                       '**EXCEPT** for `file` and `imagetyp`, which are'
#                       'always included')
#    parser.add_argument('--no-default', action='store_true',
#                        help=no_default_help)

    list_help = 'Print default list keywords put into table and exit'
    parser.add_argument('-l', '--list-default', action='store_true',
                        help=list_help)

    all_keys_help = ('Construct table from all FITS keywords present in '
                     'headers and the list of default keywords.')
    parser.add_argument('-a', '--all', action='store_true',
                        help=all_keys_help)

    default_names = DefaultFileNames()
    output_file_help = 'Name of file in which table is saved; default is '
    output_file_help += default_names.output_table
    parser.add_argument('-t', '--table-name',
                        default=default_names.output_table,
                        help=output_file_help)

    needs_object_help = 'Name of file to which list of files that need '
    needs_object_help += 'object name is saved; default is '
    needs_object_help += default_names.object_file_name
    parser.add_argument('-o', '--object-needed-list',
                        default=default_names.object_file_name,
                        help=needs_object_help)

    needs_pointing_help = 'Name of file to which list of files that need '
    needs_pointing_help += 'pointing name is saved; default is '
    needs_pointing_help += default_names.pointing_file_name
    parser.add_argument('-p', '--pointing-needed-list',
                        default=default_names.pointing_file_name,
                        help=needs_pointing_help)

    needs_filter_help = 'Name of file to which list of files that need '
    needs_filter_help += 'filter is saved; default is '
    needs_filter_help += default_names.filter_file_name
    parser.add_argument('-f', '--filter-needed-list',
                        default=default_names.filter_file_name,
                        help=needs_filter_help)

    needs_astrometry_help = 'Name of file to which list of files that need '
    needs_astrometry_help += 'astrometry is saved; default is '
    needs_astrometry_help += default_names.astrometry_file_name
    parser.add_argument('-y', '--astrometry-needed-list',
                        default=default_names.astrometry_file_name,
                        help=needs_astrometry_help)

    return parser

DEFAULT_KEYS = ['imagetyp', 'filter', 'exptime', 'ccd-temp',
                'object', 'observer', 'airmass', 'instrume',
                'RA', 'Dec', 'date-obs', 'jd', 'wcsaxes']


def main(arglist=None):
    """See script_helpers._main_function_docstring for actual documentation
    """
    parser = construct_parser()
    args = parser.parse_args(arglist)
    logger.debug('args are %s', vars(args))

    script_helpers.setup_logging(logger, args, screen_handler)

    add_file_handlers(logger, os.getcwd(), 'run_triage')

    use_keys = list(DEFAULT_KEYS)  # force a copy so DEFAULT_KEYS not modified

    if args.list_default:
        print('Keys included by default are:\n')
        keys_print = [key.upper() for key in use_keys]
        print(', '.join(keys_print))
        return use_keys

    use_keys.extend(args.key)

    if not args.dir:
        parser.error('No directory specified')

    logger.debug('use_keys are %s', use_keys)
    do_not_log_in_destination = \
        script_helpers.handle_destination_dir_logging_check(args)
    triage_directories(args.dir, keywords=use_keys,
                       all_keywords=args.all,
                       object_file_name=args.object_needed_list,
                       pointing_file_name=args.pointing_needed_list,
                       filter_file_name=args.filter_needed_list,
                       astrometry_file_name=args.astrometry_needed_list,
                       output_table=args.table_name,
                       destination=args.destination_dir,
                       no_log_destination=do_not_log_in_destination)

main.__doc__ = script_helpers._main_function_docstring(__name__)



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

from .image_collection import ImageFileCollection
from .image import ImageWithWCS
from .table_tree import TableTree

from .version import version as __version__



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

from ..feder import Feder, ApogeeAltaU9


def test_apogee_has_overscan():
    feder_obj = Feder()
    apogee_alta = feder_obj.instruments["Apogee Alta"]
    assert (apogee_alta.has_overscan([3085, 2048]))
    assert not (apogee_alta.has_overscan([3073, 2048]))


def test_apogee_fits_names():
    feder_obj = Feder()
    assert isinstance(feder_obj.instruments["Apogee Alta"],
                      ApogeeAltaU9)
    assert isinstance(feder_obj.instruments["Apogee USB/Net"],
                      ApogeeAltaU9)



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

from os import path


def get_data_dir():
    """
    Return the absolute path to the directory containing the data for testing

    This is assumed to be the subdirectory named `data` of the directory
    containing this file.

    Returns
    -------
    str
        Name of path
    """
    data_name = 'data'
    this_dir = path.dirname(path.abspath(__file__))
    data_dir = path.join(this_dir, data_name)
    if not path.isdir(data_dir):
        raise RuntimeError('No data file found in {0}'.format(this_dir))
    return data_dir



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

from tempfile import mkdtemp
import os
from shutil import rmtree
import gzip
from socket import timeout

import numpy as np
import pytest
from astropy.io import fits
from astropy.coordinates import SkyCoord, name_resolve
from ccdproc.utils.slices import slice_from_string

try:
    from .header_processing.patchers import IRAF_image_type
except ImportError:
    pass


@pytest.fixture
def triage_setup(request):
    n_test = {'files': 0, 'need_object': 0,
              'need_filter': 0, 'bias': 0,
              'compressed': 0, 'light': 0,
              'need_pointing': 0}

    test_dir = ''

    for key in n_test.keys():
        n_test[key] = 0

    test_dir = mkdtemp()
    original_dir = os.getcwd()
    os.chdir(test_dir)
    img = np.uint16(np.arange(100))

    no_filter_no_object = fits.PrimaryHDU(img)
    no_filter_no_object.header['imagetyp'] = IRAF_image_type('light')
    no_filter_no_object.writeto('no_filter_no_object_light.fit')
    n_test['files'] += 1
    n_test['need_object'] += 1
    n_test['need_filter'] += 1
    n_test['light'] += 1
    n_test['need_pointing'] += 1

    no_filter_no_object.header['imagetyp'] = IRAF_image_type('bias')
    no_filter_no_object.writeto('no_filter_no_object_bias.fit')
    n_test['files'] += 1
    n_test['bias'] += 1

    filter_no_object = fits.PrimaryHDU(img)
    filter_no_object.header['imagetyp'] = IRAF_image_type('light')
    filter_no_object.header['filter'] = 'R'
    filter_no_object.writeto('filter_no_object_light.fit')
    n_test['files'] += 1
    n_test['need_object'] += 1
    n_test['light'] += 1
    n_test['need_pointing'] += 1

    filter_no_object.header['imagetyp'] = IRAF_image_type('bias')
    filter_no_object.writeto('filter_no_object_bias.fit')
    n_test['files'] += 1
    n_test['bias'] += 1

    filter_object = fits.PrimaryHDU(img)
    filter_object.header['imagetyp'] = IRAF_image_type('light')
    filter_object.header['filter'] = 'R'
    filter_object.header['OBJCTRA'] = '00:00:00'
    filter_object.header['OBJCTDEC'] = '00:00:00'
    filter_object.writeto('filter_object_light.fit')
    n_test['files'] += 1
    n_test['light'] += 1
    n_test['need_object'] += 1
    n_test['need_pointing'] += 1

    with open('filter_object_light.fit', 'rb') as f_in:
        with gzip.open('filter_object_light.fit.gz', 'wb') as f_out:
            f_out.write(f_in.read())
    n_test['files'] += 1
    n_test['compressed'] += 1
    n_test['light'] += 1
    n_test['need_object'] += 1
    n_test['need_pointing'] += 1

    filter_object.header['RA'] = filter_object.header['OBJCTRA']
    filter_object.header['Dec'] = filter_object.header['OBJCTDEC']
    filter_object.writeto('filter_object_RA_keyword_light.fit')
    n_test['files'] += 1
    n_test['light'] += 1
    n_test['need_object'] += 1
    n_test['need_pointing'] += 1

    def teardown():
        for key in n_test.keys():
            n_test[key] = 0
        rmtree(test_dir)
        os.chdir(original_dir)
    request.addfinalizer(teardown)

    class Result(object):
        def __init__(self, n, directory):
            self.n_test = n
            self.test_dir = directory
    return Result(n_test, test_dir)


@pytest.fixture
def make_overscan_test_files(request, tmpdir):
    """
    Creates two files, one with overscan, one without for Alta U9

    Parameters

    test_dir: str
        Directory in which to create the overscan files.

    Returns

    info: list
        (working_dir, has_oscan, has_no_oscan)
        working_dir: str
            subdirectory of test_dir in which files are created
        has_oscan: str
            Name of FITS file that has overscan region
        has_no_oscan: str
            Name of FITS file that has no overscan region
    """
    from .header_processing.feder import ApogeeAltaU9, MaximDL5
    from os import path, mkdir
    import astropy.io.fits as fits
    import numpy as np

    test_dir = tmpdir
    oscan_names = ['yes_scan', 'no_scan']
    oscan = {'yes_scan': True,
             'no_scan': False}
    apogee = ApogeeAltaU9()
    working_dir = 'overscan_test'
    working_path = test_dir.mkdir(working_dir)
    add_instrument = lambda hdr: hdr.set('instrume', 'Apogee Alta')
    name_fits = lambda name: name + '.fit'
    for name in oscan_names:
        if oscan[name]:
            data = np.zeros([apogee.rows, apogee.columns])
            has_oscan = name
        else:
            osc_slice = slice_from_string(apogee.trim_region)
            data = np.zeros([apogee.rows, osc_slice[0].stop])
            no_oscan = name
        hdu = fits.PrimaryHDU(data)
        hdr = hdu.header
        add_instrument(hdu.header)
        mdl5 = MaximDL5()
        # all headers need a software name
        hdr[mdl5.fits_keyword] = mdl5.fits_name[0]
        hdr['imagetyp'] = 'LIGHT'
        hdu.writeto(path.join(working_path.strpath, name_fits(name)))

    return (working_path.strpath, name_fits(has_oscan), name_fits(no_oscan))


@pytest.fixture
def simbad_down():
    simbad_down = False
    try:
        SkyCoord.from_name("m101")
    except (name_resolve.NameResolveError, timeout):
        simbad_down = True
    return simbad_down



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import pytest
from astropy.io.fits.hdu import PrimaryHDU
from astropy.io.fits import Header

from ..fitskeyword import FITSKeyword


class TestGoodFITSKeyword(object):

    def setup_method(self, method):
        self.name = "kwd"
        self.value = 12
        self.comment = 'This is a comment'
        self.synonyms = ['kwdalt1', 'kwdalt2']
        self.keyword = FITSKeyword(name=self.name, value=self.value,
                                   comment=self.comment,
                                   synonyms=self.synonyms)
        self.hdu = PrimaryHDU()

    def test_name_setter(self):
        assert self.keyword.name == self.name.upper()

    def test_synonym_setter_length(self):
        assert len(self.keyword.synonyms) == 2

    def test_synonym_setter_values(self):
        synonyms_should_be = set([syn.upper() for syn in self.synonyms])
        assert(not synonyms_should_be.difference(self.keyword.synonyms))

    def test_names(self):
        assert (self.keyword.names == [self.keyword.name,
                self.keyword.synonyms[0], self.keyword.synonyms[1]])

    def test_history_comment(self):
        assert (self.keyword.history_comment() ==
                "Updated keyword KWD to value 12")
        assert(self.keyword.history_comment(with_name=self.synonyms[0])
               == "Updated keyword KWDALT1 to value 12")

    def test_add_header(self):
        self.keyword.add_to_header(self.hdu)
        assert self.hdu.header[self.keyword.name] == self.keyword.value
        for synonym in self.keyword.synonyms:
            assert self.hdu.header[synonym] == self.keyword.value

    def test_add_header_history(self):
        self.keyword.add_to_header(self.hdu, history=True)
        print(self.hdu.header['history'])
        assert (len(self.hdu.header['history']) ==
                (1 + len(self.keyword.synonyms)))

    def test_add_header_no_synonyms(self):
        clean_hdu = PrimaryHDU()
        self.keyword.add_to_header(clean_hdu, with_synonyms=False)
        for synonym in self.keyword.synonyms:
            try:
                clean_hdu.header[synonym]
                assert False
            except KeyError:
                assert True

    def test_add_header_from_header(self):
        clean_hdu = PrimaryHDU()
        self.keyword.add_to_header(clean_hdu.header)
        for name in self.keyword.names:
            assert clean_hdu.header[name] == self.keyword.value

    def test_handling_of_duplicate_synonyms(self):
        # should fail if duplicate synonyms are not removed in initialization
        bad_synonyms = ['bad', 'bad']
        k = FITSKeyword(name='good', synonyms=bad_synonyms)
        assert(len(k.synonyms) == len(set(bad_synonyms)))

    def test_handling_of_synonym_that_duplicates_name(self):
        name = 'bob'
        k = FITSKeyword(name=name, synonyms=name)
        assert(len(k.names) == 1)
        good_synonyms = ['bobby', 'robert']
        k = FITSKeyword(name=name, synonyms=[name] + good_synonyms)
        assert(len(k.names) == len(good_synonyms) + 1)

    def test_setting_name_to_one_of_synonyms(self):
        # doing so should remove that new name as a synonym
        k = FITSKeyword(name=self.name,
                        synonyms=self.synonyms)
        k.name = self.synonyms[0]
        assert(len(k.synonyms) == len(self.synonyms) - 1)

    def test_string_repesentation(self):
        # If I have no value does my string contain only my name and comment?
        k = FITSKeyword(name=self.name, comment=self.comment)
        string_k = str(k)
        assert self.name.upper() in string_k
        assert self.comment in string_k
        # Does value and synonyms appear in string?
        k = FITSKeyword(name=self.name, comment=self.comment, value=self.value,
                        synonyms=self.synonyms)
        string_k = str(k)
        assert str(self.value) in string_k
        for syn in self.synonyms:
            assert syn.upper() in string_k

    def test_error_raised_if_invalid_synonym(self):
        # Does synonym which is neither string nor list raise error?
        with pytest.raises(ValueError):
            FITSKeyword(name='adfa', synonyms=12)

    def test_bad_hdu_or_header_arg_raises_error(self):
        hdu_or_header = 'Not a header or hdu'
        with pytest.raises(ValueError):
            self.keyword.add_to_header(hdu_or_header)
        with pytest.raises(ValueError):
            self.keyword.set_value_from_header(hdu_or_header)

    def test_set_value_from_header(self):
        # Do I raise a value error if the keyword isn't found?
        with pytest.raises(ValueError):
            self.keyword.set_value_from_header(self.hdu.header)
        new_value = 3 * self.value
        self.hdu.header[self.name] = new_value
        # Did I get the new value from the hdu?
        self.keyword.set_value_from_header(self.hdu)
        assert self.keyword.value == new_value
        # reset to original value
        self.keyword.value = self.value
        # Can I get the value from the header?
        self.keyword.set_value_from_header(self.hdu.header)
        assert self.keyword.value == new_value
        # Do multiple (non-identical) values raise a value error?
        self.hdu.header[self.synonyms[0]] = 7 * new_value
        with pytest.raises(ValueError):
            self.keyword.set_value_from_header(self.hdu.header)



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import pytest

from astropy.table import Table
from astropy.extern import six

from .. import table_tree as tt


@pytest.fixture
def bare_group_class():
    return tt.TableTree


@pytest.fixture
def testing_table():
    # NEED A BETTER EXAMPLE HERE!
    a = [1, 1, 7, 5, 19]
    b = ['x', 'y', 'x', 'y', 'z']
    c = [4.1, 4.1, 4.1, 4.1, 5.2]
    index = range(0, len(a))
    tbl = Table([a, b, c, index],
                names=('a', 'b', 'c', 'index'),
                meta={'index_key': 'index',
                      'bad_group_key': 'sdafas'})
    return tbl


@pytest.fixture
def expected_tree(testing_table):
    # SO, now thinking it would be better if the leaves were always a list.
    # all assume the index is 'index'
    # if the keys are  a, b (in that order), then:
    tree_1 = { 1: {'x': [0],
                   'y': [1]},
               7: {'x': [2]},
               5: {'y': [3]},
              19: {'z': [4]}}
    # the keys are b, a (in that order) then:
    tree_2 = {'x': {1: [0],
                    7: [2]},
              'y': {1: [1],
                    5: [3]},
              'z': {19: [4]}}
    # the decimal entries in the table are subject to rounding, so use those
    # entries rather than hand entering them below.
    # if the keys are c (only) then:
    tree_3 = {testing_table['c'][0]: [0, 1, 2, 3],
              testing_table['c'][4]: [4]}
    # if keys are c and b:
    tree_4 = {testing_table['c'][0]: {'x': [0, 2],
                    'y': [1, 3]},
              testing_table['c'][4]: {'z': [4]}}
    tree_for_keys = {'a,b': tree_1,
                     'b,a': tree_2,
                     'c': tree_3,
                     'c,b': tree_4}
    return tree_for_keys


@pytest.fixture
def good_grouper(testing_table):
    group_keys = testing_table.colnames[0:2]
    index_key = testing_table.meta['index_key']
    grouper = tt.TableTree(testing_table, group_keys, index_key)
    return grouper


@pytest.fixture
def known_attributes():
    attributes = ['table', 'tree_keys', 'index_key']
    return attributes


def test_create_grouper(bare_group_class):
    assert issubclass(bare_group_class, tt.TableTree)


def test_grouper_has_docstring(bare_group_class):
    assert bare_group_class.__doc__
    assert len(bare_group_class.__doc__.strip()) > 0


def test_grouper_docsting_for_sections(bare_group_class):
    required_sections = ['Parameters', 'Attributes']
    for section in required_sections:
        assert section in bare_group_class.__doc__


def test_grouper_initializer_number_arguments(good_grouper, testing_table):
    assert isinstance(good_grouper, tt.TableTree)
    # should fail because too few arguments
    with pytest.raises(TypeError):
        print(tt.TableTree(['a']))
    # should fail because too many arguments
    with pytest.raises(TypeError):
        print(tt.TableTree(testing_table, testing_table.colnames[0:2],
                            testing_table.meta['index_key'], 5))


def test_group_initialization_with_bad_group_key(testing_table):
    group_keys = [testing_table.colnames[0],
                  testing_table.meta['bad_group_key']]
    index_key = testing_table.meta['index_key']
    with pytest.raises(KeyError):
        bad_grouper = tt.TableTree(testing_table, group_keys, index_key)


def test_group_initialization_checks_input_types(testing_table):
    group_keys = testing_table.colnames[0:2]
    index_key = testing_table.meta['index_key']
    # should error because first argument is not a table
    with pytest.raises(TypeError):
        bad_grouper = tt.TableTree(group_keys, group_keys, index_key)
    # should error because second argument is a string, which is iterable but
    # not a list
    with pytest.raises(TypeError):
        bad_grouper = tt.TableTree(testing_table, index_key, index_key)
    # should error because second argument is not iterable
    with pytest.raises(TypeError):
        bad_grouper = tt.TableTree(testing_table, 5, index_key)
    # should error because third argument is not a string
    with pytest.raises(TypeError):
        bad_grouper = tt.TableTree(testing_table, group_keys, [1, 2])
    # should error because third argument is not a column in the table
    with pytest.raises(KeyError):
        bad_grouper = tt.TableTree(testing_table, group_keys,
                                    testing_table.meta['bad_group_key'])
    # should error because third argument has values that are not unique
    with pytest.raises(ValueError):
        bad_grouper = tt.TableTree(testing_table, group_keys,
                                    testing_table.colnames[0])


def test_grouper_has_attributes(good_grouper, known_attributes):
    for attribute in known_attributes:
        assert hasattr(good_grouper, attribute)


def test_grouper_attributes_are_not_settable(good_grouper, known_attributes):
    dummy_value = 5
    for attribute in known_attributes:
        with pytest.raises(AttributeError):
            setattr(good_grouper, attribute, dummy_value)


def test_grouper_tree_type(good_grouper):
    assert isinstance(good_grouper, tt.RecursiveTree)


def test_grouper_tree_manual_key_addition(good_grouper):
    good_grouper['one']['two']
    assert 'one' in good_grouper.keys()
    assert 'two' in good_grouper['one'].keys()
    new_keys = range(4)
    good_grouper.add_keys(new_keys)
    for key in new_keys:
        assert key in good_grouper.keys()
        good_grouper = good_grouper[key]


def compare_trees(tree1, tree2):
    """
    Recursively compare two trees

    This only terminates if the leaves of the trees are instances of a list. In
    other words, it works fine for testing the trees created by TableTree but
    isn't really a general tree comparer.
    """
    for key in tree1.keys():
        print(key)
        assert key in tree2.keys()
        if isinstance(tree1[key], list):
            print(tree1[key])
            assert tree1[key] == tree2[key]
        else:
            print('Calling compare_trees recursively')
            compare_trees(tree1[key], tree2[key])


def test_grouper_with_expected_trees(testing_table, expected_tree):
    index_with = testing_table.meta['index_key']
    for grouping_string, good_tree in six.iteritems(expected_tree):
        grouping_keys = grouping_string.split(',')
        grouper = tt.TableTree(testing_table, grouping_keys, index_with)
        compare_trees(grouper, good_tree)


def validate_walk(grouper, good_tree):
    for parents, children, index in grouper.walk():
        working_tree = dict(good_tree)
        # walk the comparison tree down to the same level as the grouper tree
        print(parents)
        for parent in parents:
            print(parent)
            working_tree = working_tree[parent]
        if children:
            assert set(children) == set(working_tree.keys())
        if index:
            assert set(index) == set(working_tree)


def test_grouper_walk(testing_table, expected_tree):
    index_with = testing_table.meta['index_key']
    for grouping_string, good_tree in six.iteritems(expected_tree):
        grouping_keys = grouping_string.split(',')
        grouper = tt.TableTree(testing_table, grouping_keys, index_with)
        validate_walk(grouper, good_tree)



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import os
import argparse
import subprocess

from . import script_helpers
from .run_triage import DefaultFileNames
from .. import ImageFileCollection


def construct_parser():
    parser = argparse.ArgumentParser()

    source_root_help = ('All directories below this one that '
                        'contain images will be processed')
    parser.add_argument('source_root',
                        help=source_root_help,
                        nargs=1)

    group = parser.add_mutually_exclusive_group(required=True)
    dest_root_help = ('If set, image directories below ``source-root`` '
                      'will be copied into this directory tree. Only '
                      'directories that contain image files will be copied; '
                      'any intermediary directories required to contain '
                      'directories that contain images will also be created.')
    group.add_argument('--dest-root',
                       help=dest_root_help,
                       nargs=1)

    group.add_argument('--overwrite-source',
                       help=('This flag must be used to overwrite images in '
                             'the course directory.'),
                       action='store_true'
                       )

    parser.add_argument('--scripts-only',
                        help=('This script will write a single shell script '
                              'with the name provided in this option. No '
                              'images will be modified or directories '
                              'created, but the script can be run to do '
                              'those things.'),
                        action='store_true')

    parser.add_argument('-r', '--run-only',
                        help=('Select which scripts you want to run. This can '
                              'be any combination of [p]atch, [a]strometry '
                              'and [t]riage.'),
                        choices='atp')

    parser.add_argument('--no-blind',
                        help=('Disable astrometry for images without '
                              'pointing information '),
                        action='store_true')

    object_list_help = ('Path to or URL of file containing list (and '
                        'optionally coordinates of) objects that might be in '
                        'these files. If not provided it defaults to looking '
                        'for a file called obsinfo.txt in the directory '
                        'being processed')
    parser.add_argument('-o', '--object-list',
                        help=object_list_help,
                        default=None)

    script_helpers.add_console_output_args(parser)
    script_helpers.add_debug(parser)
    parser.add_argument('--quiet-log',
                        help=('Log only warnings (or worse) to '
                              'FILES AND CONSOLE while running scripts'),
                        action='store_true')

    return parser


def construct_command(script_name, source, destination,
                      common_arguments, additional_args=None):
    command = [script_name]
    command.extend(common_arguments)
    command.extend(['--destination-dir', destination])
    if additional_args:
        command.extend(additional_args)
    command.extend([source])
    command = [arg for arg in command if arg]
    return command


def main(arglist=None):
    """See script_helpers._main_function_docstring for actual documentation
    """
    parser = construct_parser()

    args = parser.parse_args(arglist)

    common_args = []

    verbose = '-v' if not args.quiet_log else ''
    quiet_console = '--quiet-console' if args.quiet_console else ''
    silent_console = '--silent-console' if args.silent_console else ''
    common_args.extend([verbose, quiet_console, silent_console])

    no_blind = args.no_blind

    object_list_option = []
    if args.object_list is not None:
        object_list_option = ['--object-list', args.object_list]

    source_root = args.source_root[0]
    dest_root = args.dest_root or [source_root]
    dest_root = dest_root[0]

    # check that source and destination roots are not same unless user has
    # explicitly requested overwrite
    if (dest_root == source_root) & (not args.overwrite_source):
        raise RuntimeError('Will not overwrite source unless you explicitly '
                           'use the option --overwrite-source')

    SEP_REPEAT = 20
    separator_format = '#' * SEP_REPEAT + ' {} ' + '#' * SEP_REPEAT

    SCRIPT_NAME = 'header_process_script.sh'

    scripts_to_run = args.run_only or 'pat'
    patch = 'p' in scripts_to_run
    astrometry = 'a' in scripts_to_run
    triage = 't' in scripts_to_run

    for root, dirs, files in os.walk(source_root):

        if not files:
            continue

        # I should probably make sure there are fits files in here.
        # Not going to...

        # Oh fine, I will test.
        fits_collection = ImageFileCollection(root)
        if not fits_collection.files:
            continue

        source_rel_to_root = os.path.relpath(root, source_root)
        destination = os.path.join(dest_root, source_rel_to_root)

        # Construct strings that will be used to actually do the processing
        make_destination = ['mkdir', '-p', destination]

        run_patch = construct_command('run_patch.py', root, destination,
                                      common_args,
                                      additional_args=object_list_option)
        if not patch:
            run_patch = ''

        # NOTE: after header patching all other scripts should use as their
        # input the directory that was the destination for run_patch if
        # run_patch was executed

        if patch:
            source_for_rest = destination
        else:
            source_for_rest = root

        if no_blind:
            additional_args = None
        else:
            additional_args = ['--blind']

        run_astrometry = construct_command('run_astrometry.py',
                                           source_for_rest,
                                           destination, common_args,
                                           additional_args=additional_args)
        if not astrometry:
            run_astrometry = ''

        run_triage = construct_command('run_triage.py', source_for_rest,
                                       destination, common_args,
                                       additional_args=['--all'])

        if not triage:
            run_triage = ''

        cmd_start_str = 'START commands for {}'.format(root)
        cmd_list = [separator_format.format(cmd_start_str)]
        for cmd in [make_destination, run_patch, run_astrometry, run_triage]:
            cmd_list.append(' '.join(cmd))
        # Re-run patch and triage if any files are missing pointing-related
        # keywords.
        if run_patch or run_triage:
            pointing_file = DefaultFileNames().pointing_file_name
            rerun_patch = construct_command('run_patch.py', source_for_rest,
                                            destination, common_args,
                                            additional_args=object_list_option)
            cmd_list.append('if [[ -e {} ]]; then'.format(pointing_file))
            cmd_list.append('    ' + ' '.join(rerun_patch))
            cmd_list.append('    ' + ' '.join(run_triage))
            cmd_list.append('fi')
        end_str = separator_format.format('END commands for {}'.format(root))
        cmd_list.append(end_str)
        cmd_list = '\n'.join(cmd_list) + '\n'*5
        with open(SCRIPT_NAME, mode='a') as f:
            f.write(cmd_list)

        if not args.scripts_only:
            subprocess.call(make_destination)
            script_path = os.path.join(destination, SCRIPT_NAME)
            with open(script_path, 'wt') as script_to_reproduce_this:
                script_to_reproduce_this.write(cmd_list)
            if patch:
                subprocess.call(run_patch)
            if astrometry:
                subprocess.call(run_astrometry)
            if triage:
                subprocess.call(run_triage)

main.__doc__ = script_helpers._main_function_docstring(__name__)



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

from collections import Iterable

from astropy.table import Table
from astropy.extern.six.moves import zip as izip
from astropy.extern import six

__all__ = ['TableTree', 'RecursiveTree']


class RecursiveTree(dict):
    """A dict-base recursive tree."""
    def __init__(self):
        super(RecursiveTree, self).__init__()

    def __missing__(self, key):
        """
        Add the method that makes this tree recursive

        When an unknown key is encountered its default value is an empty
        instance of a RecursiveTree.

        Idea is from `Stack Overflow
        <http://stackoverflow.com/questions/6780952/how-to-change-behavior-of-dict-for-an-instance>`_
        """
        value = self[key] = type(self)()
        return value

    def add_keys(self, keys, value=None):
        for key in keys[:-1]:
            self = self[key]
        if value is not None:
            self[keys[-1]] = value
        else:
            self = self[keys[-1]]


class TableTree(RecursiveTree):
    """
    Base class for grouping images hierarchically into a tree based on metadata.

    Parameters
    ----------
    table : astropy.table.Table instance
        Table containing the metadata to be used for grouping images.
    tree_keys : list of str
        Keys to be used in grouping images. Each key must be the name of a
        column in `table`.
    index_key : str
        Key which is used to indicate which rows of the input table are in each
        group; it must be the name of one of the columns in `table`. Values of
        the index must uniquely identify rows of the table (in database
        parlance, index must be able to serve as a primary key for the table).

    Attributes
    ----------
    table
    tree_keys
    index_key

    Raises
    ------
    TypeError
        Raised if the number of initialization arguments is incorrect or the
        types of any of the arguments is incorrect.
    """
    def __init__(self, *args):
        super(TableTree, self).__init__()
        if not args:
            return

        if len(args) != 3:
            raise TypeError("TableTree must be initialized with three "
                            "arguments")
        table = args[0]
        tree_keys = args[1]
        index_key = args[2]

        if not isinstance(table, Table):
            raise TypeError('First argument must be an '
                            'astropy.table.Table instance')
        if (isinstance(tree_keys, six.string_types) or
                not isinstance(tree_keys, Iterable)):
            raise TypeError('Second argument must be list-like but not '
                            'a single string.')
        if not isinstance(index_key, six.string_types):
            raise TypeError('Third argument must be a string.')
        self._table = table
        self._tree_keys = tree_keys
        self._validate_group_keys()
        self._index_key = index_key
        self._validate_index()
        self._build_tree()

    def _validate_group_keys(self):
        """
        Each key must be a table column name.
        """
        for key in self._tree_keys:
            try:
                self._table[key]
            except KeyError:
                raise

    def _validate_index(self):
        """
        The rows in the `table` column named `index` must be unique and `index`
        must be a column in the table. The second case is raised automatically
        by astropy.table.Table
        """
        index_column = self._table[self._index_key]
        if len(set(index_column)) != len(self._table):
            raise ValueError('The table column named {0} cannot be used as '
                             'and index because its values are '
                             'not unique'.format(self._index_key))

    def _build_tree(self):
        """
        Construct tree from grouping done by astropy table

        Because the tree is constructed
        """
        grouped = self.table.group_by(self.tree_keys)
        columns = list(self.tree_keys)
        columns.append(self.index_key)
        for group_key, members in izip(grouped.groups.keys,
                                       grouped.groups):
            key_list = list(group_key)  # table rows can't be indexed w/slice
            indexes = list(members[self.index_key])
            self.add_keys(key_list, value=indexes)

    @property
    def table(self):
        """
        astropy.table.Table of metadata used to group rows.
        """
        return self._table

    @property
    def tree_keys(self):
        """
        list of str, Table columns to be used in grouping the rows.
        """
        return self._tree_keys

    @property
    def index_key(self):
        """
        str, Name of column whose values uniquely identify each row.
        """
        return self._index_key

    def walk(self, *args, **kwd):
        """
        Walk the grouped tree

        The functionality provided is similar to that in os.walk: starting at
        the top of tree, yield a tuple of return values indicating parents,
        children and rows at each level of the tree.

        Parameters
        ----------
        None

        Returns
        -------
        parents, children, index : lists
        parents : list
            Dictionary keys that led to this return
        children : list
            Child nodes at this level
        index : list
            Index values for the items in the table that correspond to the
            values in `parents`
        """

        parent = kwd.pop("parent", [])
        if args:
            use_dict = args[0]
        else:
            use_dict = self

        try:
            tree_nodes = use_dict.keys()
            indexes = []
            yield parent, tree_nodes, indexes
        except AttributeError:
            tree_nodes = []
            yield parent, tree_nodes, use_dict

        for node in tree_nodes:
            new_parent = list(parent)
            new_parent.append(node)
            for val in self.walk(use_dict[node], parent=new_parent):
                yield val



"""
DESCRIPTION
-----------

For each directory provided on the command line the
headers all of the FITS files in that directory are modified
to add information like LST, apparent object position, and more.
See the full documentation for a list of the specific keywords
that are modified.

Header patching
^^^^^^^^^^^^^^^

This is basically a wrapper around the function
:func:`patch_headers.patch_headers` with the options set so that:

    + "Bad" keywords written by MaxImDL 5 are purged.
    + ``IMAGETYP`` keyword is changed from default MaxIM DL style
      to IRAF style (e.g. "Bias Frame" to "BIAS")
    + Additional useful times like LST, JD are added to the header.
    + Apparent position (Alt/Az, hour angle) are added to the header.
    + Information about overscan is added to the header.
    + Files are overwritten.

For more control over what is patched and where the patched files are saved
see the documentation for ``patch_headers`` at
:func:`patch_headers.patch_headers`.

Adding OBJECT keyword
^^^^^^^^^^^^^^^^^^^^^

``run_patch`` also adds the name of the object being observed when
appropriate (i.e. only for light files) and possible. It needs to be
given a list of objects; looking up the coordinates for those objects
requires an Internet connection. See

For a detailed description of the object list file see
:func:`Object file format <patch_headers.read_object_list>`.

for a detailed description of the function that actually adds the object name
see :func:`patch_headers.add_object_info`.

If no object list is specified or present in the directory being processed
the `OBJECT` keyword is simply not added to the FITS header.

.. Note::
    This script is **NOT RECURSIVE**; it will not process files in
    subdirectories of the the directories supplied on the command line.

.. WARNING::
    This script OVERWRITES the image files in the directories
    specified on the command line unless you use the --destination-dir
    option.

EXAMPLES
--------

Invoking this script from the command line::

    run_patch.py /my/folder/of/images

To work on the same folder from within python, do this::

    from msumastro.scripts import run_patch
    run_patch.main(['/my/folder/of/images'])

To use the same object list for several different directories do this::

    run_patch.py --object-list path/to/list.txt dir1 dir2 dir3

where ``path/to/list.txt`` is the path to your object list and ``dir1``,
``dir2``, etc. are the directories you want to process.

From within python this would be::

    from msumastro.scripts import run_patch
    run_patch.main(['--object-list', 'path/to/list.txt',
                   'dir1', 'dir2', 'dir3'])
"""
from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

from os import getcwd
from os import path
import warnings
import logging

from ..header_processing import patch_headers, add_object_info, list_name_is_url
from ..customlogger import console_handler, add_file_handlers
from .script_helpers import (setup_logging, construct_default_parser,
                             handle_destination_dir_logging_check,
                             _main_function_docstring)

logger = logging.getLogger()
screen_handler = console_handler()
logger.addHandler(screen_handler)

DEFAULT_OBJ_LIST = 'obsinfo.txt'
DEFAULT_OBJECT_URL = ('https://raw.github.com/mwcraig/feder-object-list'
                      '/master/feder_object_list.csv')


def patch_directories(directories, verbose=False, object_list=None,
                      destination=None,
                      no_log_destination=False,
                      overscan_only=False,
                      script_name='run_patch'):
    """
    Patch all of the files in each of a list of directories.

    Parameters
    ----------

    directories : str or list of str
        Directory or directories whose FITS files are to be processed.

    verbose : bool, optional
        Control amount of logging output.

    object_list : str, optional
        Path to or URL of a file containing a list of objects that
        might be in the files in `directory`. If not provided it defaults
        to looking for a file called `obsinfo.txt` in the directory being
        processed.

    destination : str, optional
        Path to directory in which patched images will be stored. Default
        value is None, which means that **files will be overwritten** in
        the directory being processed.
    """
    no_explicit_object_list = (object_list is None)
    if not no_explicit_object_list:
        if list_name_is_url(object_list):
            obj_dir = None
            obj_name = object_list
        else:
            full_path = path.abspath(object_list)
            obj_dir, obj_name = path.split(full_path)

    for currentDir in directories:
        if destination is not None:
            working_dir = destination
        else:
            working_dir = currentDir

        if (not no_log_destination) and (destination is not None):
            add_file_handlers(logger, working_dir, 'run_patch')

        logger.info("Working on directory: %s", currentDir)

        with warnings.catch_warnings():
            # suppress warning from overwriting FITS files
            ignore_from = 'astropy.io.fits.hdu.hdulist'
            warnings.filterwarnings('ignore', module=ignore_from)
            if overscan_only:
                patch_headers(currentDir, new_file_ext='', overwrite=True,
                              save_location=destination,
                              purge_bad=False,
                              add_time=False,
                              add_apparent_pos=False,
                              add_overscan=True,
                              fix_imagetype=False,
                              add_unit=False)

            else:
                patch_headers(currentDir, new_file_ext='', overwrite=True,
                              save_location=destination)

                default_object_list_present = path.exists(path.join(currentDir,
                                                          DEFAULT_OBJ_LIST))
                if (default_object_list_present and no_explicit_object_list):
                    obj_dir = currentDir
                    obj_name = DEFAULT_OBJ_LIST
                add_object_info(working_dir, new_file_ext='', overwrite=True,
                                save_location=destination,
                                object_list_dir=obj_dir, object_list=obj_name)


def construct_parser():
    parser = construct_default_parser(__doc__)

    object_list_help = ('Path to or URL of file containing list (and '
                        'optionally coordinates of) objects that might be in '
                        'these files. If not provided it defaults to looking '
                        'for a file called obsinfo.txt in the directory '
                        'being processed')
    parser.add_argument('-o', '--object-list',
                        help=object_list_help,
                        default=DEFAULT_OBJECT_URL)
    parser.add_argument('--overscan-only', action='store_true',
                        help='Only add appropriate overscan keywords')
    return parser


def main(arglist=None):
    """See script_helpers._main_function_docstring for actual documentation
    """
    parser = construct_parser()
    args = parser.parse_args(arglist)

    setup_logging(logger, args, screen_handler)

    add_file_handlers(logger, getcwd(), 'run_patch')

    do_not_log_in_destination = handle_destination_dir_logging_check(args)

    patch_directories(args.dir, verbose=args.verbose,
                      object_list=args.object_list,
                      destination=args.destination_dir,
                      no_log_destination=do_not_log_in_destination,
                      overscan_only=args.overscan_only)

main.__doc__ = _main_function_docstring(__name__)



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

from itertools import chain
import logging
import datetime

from astropy.coordinates import EarthLocation
import astropy.units as u
import numpy as np
from ccdproc.utils.slices import slice_from_string

from .fitskeyword import FITSKeyword

logger = logging.getLogger(__name__)

__all__ = ['FederSite', 'ImageSoftware', 'Instrument', 'ApogeeAltaU9',
           'MaximDL4', 'MaximDL5']


class FederSite(EarthLocation):
    """
    The Feder Observatory site.

    An astropy location with the observatory location pre-set to:

        + `lat` = 46.86678 degrees North
        + `long` = -96.453278 degrees East
        + `height` = 311.8 meters

    and a few additional properties/methods that are convenient:

        + `name` = Feder Observatory
    """

    def __new__(cls):
        return EarthLocation.__new__(FederSite,
                                     lat=46.86678*u.degree,
                                     lon=-96.453278*u.degree,
                                     height=311.8*u.m)

    def __init__(self):
        self._name = 'Feder Observatory'

    @property
    def name(self):
        return self._name


class Instrument(object):
    """
    Telescope instrument with simple properties.

    Parameters
    ----------
    name : str
        Name of the instrument.
    fits_names : list of str
        List of names by which the instrument is known in FITS headers
    rows : int
        Number of rows in an image produced by this instrument, including
        overscan.
    columns : int
        Number of columns in an image produced by this instrument, including
        overscan.

    image_unit : astropy.units.Unit
        Unit of the image; default value is ``None``

    trim_region : string
        Region of the CCD that should be preserved after overscan subtraction.
        Should use *FITS* conventions for specifying slices (i.e. slice
        starts at 1, includes endpoint, and uses FITS NAXIS1, NAXIS2 for
        order of indices).

    useful_overscan_region : string
        Complete specification of the region of the CCD actually useful for
        overscan calibration. This may (or may not) be smaller than the
        entire portion of the chip the manufacturer labels as overscan.
        Should use *FITS* conventions for specifying slices (i.e. slice
        starts at 1, includes endpoint, and uses FITS NAXIS1, NAXIS2 for
        order of indices).

    Examples
    --------

    Consider an image whose dimensions as given in its FITS header are
    ``NAXIS1 = 3085`` and ``NAXIS2 = 2048`` with an overscan region that
    begins at position 3073 along axis 1. The useful part of that overscan is
    from FITS column 3076 up to and including, 3079, and the full range of
    rows (``NAXIS2``). The correct overscan settings for this instrument are::

        # Note not all of the overscan region is actually useful.
        useful_overscan_region = '[3076:3079, :]'
        # But the whole overscan region should be trimmed away in reduction.
        trim_region = '[1:3073, :]'
    """

    def __init__(self, name, fits_names=None,
                 rows=0, columns=0,
                 image_unit=None,
                 trim_region=None,
                 useful_overscan_region=None):
        self.name = name
        self.fits_names = fits_names
        self.rows = rows
        self.columns = columns
        self.image_unit = image_unit
        self.trim_region = trim_region
        self.useful_overscan = useful_overscan_region

    def has_overscan(self, image_dimensions):
        """
        Determine whether an image taken by this instrument has overscan

        Parameters
        ----------
        image_dimensions : list-like with two elements
            Shape of the image; can be any type as long as it has two elements.
            The order should be the FITS order, ``NAXIS1`` then ``NAXIS2``.

        Returns
        -------
        bool
            Indicates whether or not image has overscan present.
        """
        # Grab the trim region as a slice to make it easier to access end
        # points. Do *not* convert from FITS convention because input
        # dimensions follow FITS conventions.
        trim_dim = slice_from_string(self.trim_region)

        dim_end = lambda ax: (trim_dim[ax].stop + 1
                              if trim_dim[ax].stop is not None
                              else image_dimensions[ax])
        # print(trim_dim1, trim_dim2)
        # if trim_dim1.stop is not None:
        #     dim1_end = trim_dim1.stop + 1
        # else:
        #     # None means use the whole thing....
        #     dim1_end = image_dimensions[0]

        # if trim_dim2.stop is not None:
        #     dim2_end = trim_dim2.stop + 1
        # else:
        #     # None means use the whole thing....
        #     dim2_end = image_dimensions[1]

        if (dim_end(0) < image_dimensions[0] or
            dim_end(1) < image_dimensions[1]):

            return True
        else:
            return False


class ApogeeAltaU9(Instrument):
    """
    The Apogee Alta U9
    """
    def __init__(self):
        Instrument.__init__(self, "Apogee Alta U9",
                            fits_names=["Apogee Alta", "Apogee USB/Net"],
                            rows=2048, columns=3085,
                            useful_overscan_region='[3076:3079, :]',
                            trim_region='[1:3073, :]',
                            image_unit=u.adu)


class SBIGSpectrometer(Instrument):
    """
    SBIG ST-7 spectromter.
    """
    def __init__(self):
        super(SBIGSpectrometer, self).__init__("SBIG ST-7 Spectrometer",
                                               fits_names=["SBIG ST-7"])


class ImageSoftware(object):

    """
    Represents software that takes images at telescope.

    Parameters
    ----------
    name : str
        Name of the software. Can be the same is the name in the FITS file,
        or not.
    fits_keyword : str
        Name of the FITS keyword that contains the name of the software.
    fits_name : list of str
        Name of the software as written in the FITS file
    major_version : int
        Major version number of the software.
    minor_version : int
        Minor version number of the software.
    bad_keywords : list of strings
        Names of any keywords that should be removed from the FITS before
        further processing.
    purged_flag_keyword : str, optional
        Name of the keyword which indicates whether bad keywords have already
        been purged. Default value is 'PURGED'
    """

    def __init__(self, name, fits_name=None,
                 major_version=None,
                 minor_version=None,
                 bad_keywords=None,
                 fits_keyword=None,
                 purged_flag_keyword=None):
        self.name = name
        self.fits_name = fits_name
        self.major_version = major_version
        self.minor_version = minor_version
        self.bad_keywords = bad_keywords
        self.fits_keyword = fits_keyword
        self.purged_flag_keyword = purged_flag_keyword or "PURGED"

    def created_this(self, version_string):
        """
        Indicate whether version string matches this software

        Parameters
        ----------
        version_string : str
            String from FITS header that indicates software version.

        Returns
        -------
        bool
            ``True`` if the version string matches the software instance.
        """
        return version_string in self.fits_name


class MaximDL4(ImageSoftware):

    """
    Represents MaximDL version 4, all sub-versions
    """

    def __init__(self):
        fits_name = ['MaxIm DL Version 4.10']
        super(MaximDL4, self).__init__("MaxImDL",
                                       fits_name=fits_name,
                                       major_version=4,
                                       minor_version=10,
                                       bad_keywords=['OBSERVER'],
                                       fits_keyword='SWCREATE')


class MaximDL5(ImageSoftware):

    """
    Represents MaximDL version 5, all sub-versions.

    Subversions are included by listing the FITS names of all versions that
    have been used at Feder Observatory.
    """

    def __init__(self):
        bad_keys = ['OBJECT', 'JD', 'JD-HELIO', 'OBJCTALT', 'OBJCTAZ',
                    'OBJCTHA', 'AIRMASS', 'OBSERVER']
        fits_name = ['MaxIm DL Version 5.21 130912 01A17',
                     'MaxIm DL Version 5.21 120829 2R1M0',
                     'MaxIm DL Version 5.23 130912 01A17',
                     'MaxIm DL Version 5.15']
        super(MaximDL5, self).__init__("MaxImDL",
                                       fits_name=fits_name,
                                       major_version=5,
                                       minor_version=21,
                                       bad_keywords=bad_keys,
                                       fits_keyword='SWCREATE'
                                       )


class SBIGCCDOps(ImageSoftware):
    """
    Represents software used to create images from the SBIG spectrometer.
    """
    def __init__(self):
        bad_keys = []
        fits_name = ['SBIG Win CCDOPS Version 5.47 Build 6-NT']
        super(SBIGCCDOps, self).__init__("SBIG CCDOps",
                                         fits_name=fits_name,
                                         major_version=5,
                                         minor_version=47,
                                         bad_keywords=[],
                                         fits_keyword='SWCREATE')


class Feder(object):
    """
    Class encapsulating site, instrument, and software information for Feder
    Observatory.

    Attributes
    ----------
    site : feder.FederSite instance
    instrument : dict
        Instruments available; key is name, value is an :py:class:`Instrument`
    software : dict
        Software available; key is name, value is an :class:`ImageSoftware`
    software_FITS_keywords : list of str
        FITS names of all software available.
    keywords_for_all_files
    keywords_for_light_files
    keywords_for_overscan
    """
    def __init__(self):
        self.site = FederSite()
        self._apogee_alta_u9 = ApogeeAltaU9()
        self._sbig_spectrometer = SBIGSpectrometer()
        self._instrument_objects = [self._apogee_alta_u9,
                                    self._sbig_spectrometer]
        self.instruments = {}
        for instrument in self._instrument_objects:
            for name in instrument.fits_names:
                self.instruments[name] = instrument
        self._maximdl4 = MaximDL4()
        self._maximdl5 = MaximDL5()
        self._sbig_ccdops = SBIGCCDOps()
        self._software_objects = [self._maximdl4, self._maximdl5, self._sbig_ccdops]
        self.software = {}
        self.software_FITS_keywords = []
        for software in self._software_objects:
            for name in software.fits_name:
                self.software[name] = software
                self.software_FITS_keywords.append(software.fits_keyword)
        self.software_FITS_keywords = list(set(self.software_FITS_keywords))
        self._keywords_for_all_files = []
        self._set_site_keywords_values()
        self._time_keywords_to_set()
        self._keywords_for_light_files = []
        self._define_keywords_for_light_files()
        self._overscan_keywords = []
        self._define_overscan_keywords()
        for key in chain(self.keywords_for_all_files,
                         self.keywords_for_light_files,
                         self.keywords_for_overscan):
            name = key.name
            name = name.replace('-', '_')
            setattr(self, name, key)

    @property
    def keywords_for_all_files(self):
        """
        List of :class:`~msumastro.header_processing.fitskeyword.FITSKeyword` s
        whose values need to be set for all image types.
        """
        return self._keywords_for_all_files

    @property
    def keywords_for_light_files(self):
        """
        List of :class:`~msumastro.header_processing.fitskeyword.FITSKeyword` s
        whose values need to be set only for light image types.
        """
        return self._keywords_for_light_files

    @property
    def keywords_for_overscan(self):
        """
        List of :class:`~msumastro.header_processing.fitskeyword.FITSKeyword` s
        related to overscan.
        """
        return self._overscan_keywords

    def _define_keywords_for_light_files(self):
        RA = FITSKeyword(name='ra',
                         comment='Approximate RA at EQUINOX',
                         synonyms=['objctra'])

        Dec = FITSKeyword(name='DEC',
                          comment='Approximate DEC at EQUINOX',
                          synonyms=['objctdec'])

        target_object = FITSKeyword(name='object',
                                    comment='Target of the observations')

        hour_angle = FITSKeyword(name='ha',
                                 comment='Hour angle')

        airmass = FITSKeyword(name='airmass',
                              comment='Airmass (Sec(Z)) at start of observation',
                              synonyms=['secz'])

        altitude = FITSKeyword(name='alt-obj',
                               comment='[degrees] Altitude of object, no refraction')

        azimuth = FITSKeyword(name='az-obj',
                              comment='[degrees] Azimuth of object, no refraction')
        self._keywords_for_light_files.append(RA)
        self._keywords_for_light_files.append(Dec)
        self._keywords_for_light_files.append(target_object)
        self._keywords_for_light_files.append(hour_angle)
        self._keywords_for_light_files.append(airmass)
        self._keywords_for_light_files.append(altitude)
        self._keywords_for_light_files.append(azimuth)

    def _set_site_keywords_values(self):
        latitude = FITSKeyword(name="latitude",
                               comment='[degrees] Observatory latitude',
                               synonyms=['sitelat'])

        longitude = FITSKeyword(name='longitud',
                                comment='[degrees east] Observatory longitude',
                                synonyms='sitelong')
        obs_altitude = FITSKeyword(name='altitude',
                                   comment='[meters] Observatory altitude')
        lat_lon_format = {'sep': ':', 'pad': True, 'alwayssign': True}
        latitude.value = self.site.latitude.to_string(**lat_lon_format)
        longitude.value = self.site.longitude.to_string(**lat_lon_format)
        obs_altitude.value = self.site.height.value
        self._keywords_for_all_files.extend([latitude, longitude,
                                            obs_altitude])

    def _time_keywords_to_set(self):
        LST = FITSKeyword(name='LST',
                          comment='Local Sidereal Time at start of observation')

        JD = FITSKeyword(name='jd-obs',
                         comment='Julian Date at start of observation')

        MJD = FITSKeyword(name='mjd-obs',
                          comment='Modified Julian date at start of observation')
        self._keywords_for_all_files.extend([LST, JD, MJD])

    def _define_overscan_keywords(self):
        overscan_region = FITSKeyword(name='biassec',
                                    comment='Useful region of the overscan')

        trim_region = FITSKeyword(name='trimsec',
            comment=('Region to keep after trimming overscan'))

        self._overscan_keywords.extend([overscan_region,
                                       trim_region])



"""A set of functions to standardize some options for python scripts"""

from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import logging
import argparse
from os import getcwd, path

logger = logging.getLogger(__name__)


def setup_parser_help(parser, additional_docs=None):
    """
    Set formatting for parser to raw and add docstring to help output

    Parameters
    ----------

    parser : `ArgumentParser`
        The parser to be modified.

    additional_docs: str
        Any documentation to be added to the documentation produced by
        `argparse`

    """
    parser.formatter_class = argparse.RawDescriptionHelpFormatter
    if additional_docs is not None:
        parser.epilog = additional_docs


def add_verbose(parser):
    """
    Add a verbose option (--verbose or -v) to parser.

    Parameters:
    -----------

    parser : `ArgumentParser`

    """

    verbose_help = "provide more information during processing"
    parser.add_argument("-v", "--verbose", help=verbose_help,
                        action="store_true")


def add_directories(parser, nargs_in='+'):
    """
    Add a positional argument that is one or more directories.

    Parameters
    ----------

    parser : `ArgumentParser`

    """

    parser.add_argument("dir", metavar='dir', nargs=nargs_in,
                        help="Directory to process")


def add_destination_directory(parser):
    """
    Add a destination directory option

    Parameters
    ----------

    parser : `ArgumentParser`

    """
    arg_help = 'Directory in which output from this script will be stored'
    parser.add_argument("-d", "--destination-dir",
                        help=arg_help,
                        default=None)


def add_debug(parser):
    """
    Add a debug option to produce very verbose output

    Parameters
    ----------

    parser : `ArgumentParser`
    """
    arg_help = 'Turn on very detailed logging output'
    parser.add_argument('--debug', help=arg_help, action='store_true')


def add_no_log_destination(parser):
    """
    Add option to suppress logging to files in destination directory
    """
    arg_help = 'Do not write log files to destination directory'
    parser.add_argument('-n', '--no-log-destination',
                        help=arg_help, action='store_true')


def add_console_output_args(parser):
    parser.add_argument('--quiet-console',
                        help=('Log only errors (or worse) to console '
                              'while running scripts'),
                        action='store_true')
    parser.add_argument('--silent-console',
                        help=('Turn off all logging output to console'),
                        action='store_true')


def construct_default_parser(docstring=None):
    parser = argparse.ArgumentParser()
    if docstring is not None:
        setup_parser_help(parser, docstring)
    add_verbose(parser)
    add_directories(parser)
    add_destination_directory(parser)
    add_debug(parser)
    add_no_log_destination(parser)
    add_console_output_args(parser)

    return parser


def setup_logging(logger, args, screen_handler):
    logger.setLevel(logging.WARNING)
    if args.verbose:
        logger.setLevel(logging.INFO)

    if args.debug:
        logger.setLevel(logging.DEBUG)

    if args.quiet_console:
        screen_handler.setLevel(logging.WARNING)

    if args.silent_console:
        logger.removeHandler(screen_handler)


def handle_destination_dir_logging_check(args):
    """
    Perform error checking for command line arguments
    """
    # turn off destination logging if we are running in the destination
    # directory because we always create logs in the working directory...
    do_not_log_in_destination = args.no_log_destination
    cwd = getcwd()

    logger.debug('cwd: %s', cwd)
    if args.destination_dir:
        try:
            dest_is_cwd = path.samefile(args.destination_dir, cwd)
        except OSError:
            dest_is_cwd = False
    elif args.dir:
        try:
            dest_is_cwd = any([path.samefile(d, cwd) for d in args.dir])
        except OSError:
            dest_is_cwd = False
    else:
        dest_is_cwd = False

    logger.debug('dest_is_cwd: %s', dest_is_cwd)
    if dest_is_cwd:
        if do_not_log_in_destination:
            raise RuntimeError('option --no-log-destination cannot be used '
                               'when running in the destination directory '
                               'because a log is always made in the '
                               'directory in which the script is run')
        do_not_log_in_destination = True
    return do_not_log_in_destination


def _main_function_docstring(command_name):
    """
    Wrapper for invoking {} from the command line

    Parameters
    ----------

    arglist : list of strings, optional
        If set, use this arglist instead of `sys.argv` for parsing command
        line arguments. Primarily useful for testing.
    """
    return _main_function_docstring.__doc__.format(command_name)



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import fnmatch
from os import listdir, path
import logging

import numpy as np
import numpy.ma as ma

from astropy.table import Table, vstack
import astropy.io.fits as fits
from astropy.extern import six

logger = logging.getLogger(__name__)

__all__ = ['ImageFileCollection']


class ImageFileCollection(object):

    """
    Representation of a collection of image files.

    The class offers a table summarizing values of
    keywords in the FITS headers of the files in the collection and offers
    convenient methods for iterating over the files in the collection. The
    generator methods use simple filtering syntax and can automate storage
    of any FITS files modified in the loop using the generator.

    Parameters
    ----------
    location : str, optional
        path to directory containing FITS files
    keywords : list of str or '*', optional
        Keywords that should be used as column headings in the summary table.
        If the value is or includes '*' then all keywords that appear in any
        of the FITS headers of the files in the collection become table
        columns.
    info_file : str, optional
        Path to file that contains a table of information about FITS files.

    Attributes
    ----------
    location
    keywords
    files
    summary_info

    Raises
    ------

    ValueError
        Raised if keywords are set to a combination of '*' and any other
        value.
    """

    def __init__(self, location=None, keywords=None, info_file=None):
        self._location = location
        self._files = []
        if location:
            self._files = self._fits_files_in_directory()
        self._summary_info = {}
        if keywords is None:
            keywords = []
        if info_file is not None:
            try:
                info_path = path.join(self.location, info_file)
            except (AttributeError, TypeError):
                info_path = info_file
            try:
                self._summary_info = Table.read(info_path,
                                                format='ascii',
                                                delimiter=',')
                self._summary_info = Table(self._summary_info,
                                           masked=True)
            except IOError:
                if location:
                    logger.warning('Unable to open table file %s, will try '
                                   'initializing from location instead',
                                   info_path)
                else:
                    raise

        # Used internally to keep track of whether the user asked for all
        # keywords or a specific list. The keywords setter takes care of
        # actually setting the correct value, this just ensure that there
        # is always *some* value.
        self._all_keywords = False

        if keywords:
            self.keywords = keywords

    @property
    def summary(self):
        """
        astropy.table.Table of values of FITS keywords for files in the
        collection.

        Each keyword is a column heading. In addition, there is a column
        called 'file' that contains the name of the FITS file. The directory
        is not included as part of that name.
        """
        return self._summary_info

    @property
    def summary_info(self):
        """
        Deprecated -- use summary instead -- astropy.table.Table of values of
        FITS keywords for files in the collection.

        Each keyword is a column heading. In addition, there is a column
        called 'file' that contains the name of the FITS file. The directory
        is not included as part of that name.
        """
        return self._summary_info

    @property
    def location(self):
        """
        str, Path name to directory containing FITS files
        """
        return self._location

    @property
    def keywords(self):
        """
        list of str, Keywords currently in the summary table.

        Setting the keywords causes the summary table to be regenerated unless
        the new keywords are a subset of the old.
        """
        if self.summary_info:
            return self.summary_info.keys()
        else:
            return []

    @keywords.setter
    def keywords(self, keywords=None):
        # since keywords are drawn from self.summary_info, setting
        # summary_info sets the keywords.
        if keywords is None:
            self._summary_info = []
            return

        if keywords == '*':
            self._all_keywords = True
        else:
            self._all_keywords = False

        logging.debug('keywords in setter before pruning: %s', keywords)

        # remove duplicates and force a copy
        new_keys = list(set(keywords))
        logging.debug('keywords after pruning %s', new_keys)

        full_new_keys = list(set(new_keys))
        full_new_keys.append('file')
        full_new_set = set(full_new_keys)
        current_set = set(self.keywords)
        if full_new_set.issubset(current_set):
            logging.debug('table columns before trimming: %s',
                          ' '.join(current_set))
            cut_keys = current_set.difference(full_new_set)
            logging.debug('will try removing columns: %s',
                          ' '.join(cut_keys))
            for key in cut_keys:
                self._summary_info.remove_column(key)
            logging.debug('after removal column names are: %s',
                          ' '.join(self.keywords))
        else:
            logging.debug('should be building new table...')
            self._summary_info = self._fits_summary(header_keywords=new_keys)

    @property
    def files(self):
        """
        list of str, Unfiltered list of FITS files in location.
        """
        return self._files

    def values(self, keyword, unique=False):
        """
        List of values for a keyword.

        Parameters
        ----------
        keyword : str
            Keyword (i.e. table column) for which values are desired.

        unique : bool, optional
            If True, return only the unique values for the keyword.

        Returns
        -------
        list
            Values as a list.
        """
        if keyword not in self.keywords:
            raise ValueError(
                'keyword %s is not in the current summary' % keyword)

        if unique:
            return list(set(self.summary_info[keyword]))
        else:
            return list(self.summary_info[keyword])

    def files_filtered(self, **kwd):
        """Determine files whose keywords have listed values.

        `**kwd` is list of keywords and values the files must have.

        The value '*' represents any value.
         A missing keyword is indicated by value ''

        Example:
        >>> keys = ['imagetyp','filter']
        >>> collection = ImageFileCollection('test/data', keywords=keys)
        >>> collection.files_filtered(imagetyp='LIGHT', filter='R')
        >>> collection.files_filtered(imagetyp='*', filter='')

        NOTE: Value comparison is case *insensitive* for strings.
        """
        # force a copy by explicitly converting to a list
        current_file_mask = list(self.summary_info['file'].mask)
        self._find_keywords_by_values(**kwd)
        filtered_files = self.summary_info['file'].compressed()
        self.summary_info['file'].mask = current_file_mask
        return filtered_files

    def refresh(self):
        """
        Refresh the collection by re-reading headers.
        """
        keywords = '*' if self._all_keywords else self.keywords
        # Re-load list of files
        self._files = self._fits_files_in_directory()
        self._summary_info = self._fits_summary(header_keywords=keywords)

    def _dict_from_fits_header(self, file_name, input_summary=None,
                               missing_marker=None):
        """
        Construct a dictionary whose keys are the header keywords and values
        are a list of the values from this file and the input dictionary.

        Parameters
        ----------

        file_name : str
            Name of FITS file.

        input_summary : dict
            Existing dictionary to which new values should be appended.

        Returns
        -------

        file_table : astropy.table.Table
        """
        from collections import OrderedDict

        def _add_val_to_dict(key, value, tbl_dict, n_prev):
            try:
                tbl_dict[key.lower()].append(value)
            except KeyError:
                tbl_dict[key.lower()] = [missing_marker] * n_previous
                tbl_dict[key.lower()].append(value)

        if input_summary is None:
            summary = OrderedDict()
            n_previous = 0
        else:
            summary = input_summary
            n_previous = len(summary['file'])

        h = fits.getheader(file_name)
        assert 'file' not in h

        # Try opening header before this so that file name is only added if
        # file is valid FITS
        try:
            summary['file'].append(path.basename(file_name))
        except KeyError:
            summary['file'] = [path.basename(file_name)]

        missing_in_this_file = [k for k in summary if (k not in h and
                                                       k != 'file')]

        multi_entry_keys = {'comment': [],
                            'history': []}

        for k, v in six.iteritems(h):
            if k == '':
                continue

            if k.lower() in ['comment', 'history']:
                multi_entry_keys[k.lower()].append(str(v))
                # Accumulate these in a separate dictionary until the
                # end to avoid adding multiple entries to summary.
                continue
            else:
                val = v

            _add_val_to_dict(k, val, summary, n_previous)

        for k, v in six.iteritems(multi_entry_keys):
            if v:
                joined = ','.join(v)
                _add_val_to_dict(k, joined, summary, n_previous)

        for missing in missing_in_this_file:
            summary[missing].append(missing_marker)

        return summary

    def _set_column_name_case_to_match_keywords(self, header_keys,
                                                summary_table):
        key_name_dict = {k.lower(): k for k in header_keys
                         if k != k.lower()}

        for lcase, user_case in six.iteritems(key_name_dict):
            try:
                summary_table.rename_column(lcase, user_case)
            except KeyError:
                pass

    def _fits_summary(self, header_keywords=None):
        """
        Generate a summary table of keywords from FITS headers.

        Parameters
        ----------

        header_keywords : list of str or '*', optional
            Keywords whose value should be extracted from FITS headers.
            Default value is ``None``.
        """
        from astropy.table import MaskedColumn

        if not self.files:
            return None

        # Get rid of any duplicate keywords, also forces a copy.
        header_keys = set(header_keywords)
        header_keys.add('file')

        file_name_column = MaskedColumn(name='file', data=self.files)

        if not header_keys or (header_keys == set(['file'])):
            summary_table = Table(masked=True)
            summary_table.add_column(file_name_column)
            return summary_table

        summary_dict = None
        missing_marker = None

        for file_name in file_name_column:
            file_path = path.join(self.location, file_name)
            try:
                summary_dict = self._dict_from_fits_header(
                    file_path, input_summary=summary_dict,
                    missing_marker=missing_marker)
            except IOError as e:
                logger.warning('Unable to get FITS header for file %s: %s',
                               file_path, e)
                continue

        summary_table = Table(summary_dict, masked=True)

        for column in summary_table.colnames:
            summary_table[column].mask = [v is missing_marker
                                          for v in summary_table[column]]

        self._set_column_name_case_to_match_keywords(header_keys,
                                                     summary_table)
        missing_columns = header_keys - set(summary_table.colnames)
        missing_columns -= set(['*'])

        length = len(summary_table)
        for column in missing_columns:
            all_masked = MaskedColumn(name=column, data=np.zeros(length),
                                      mask=np.ones(length))
            summary_table.add_column(all_masked)

        if '*' not in header_keys:
            summary_table.keep_columns(header_keys)

        if not summary_table.masked:
            summary_table = Table(summary_table, masked=True)

        return summary_table

    def _find_keywords_by_values(self, **kwd):
        """
        Find files whose keywords have given values.

        `**kwd` is list of keywords and values the files must have.

        The value '*' represents any value.
         A missing keyword is indicated by value ''

        Example:
        >>> keys = ['imagetyp','filter']
        >>> collection = ImageFileCollection('test/data', keywords=keys)
        >>> collection.files_filtered(imagetyp='LIGHT', filter='R')
        >>> collection.files_filtered(imagetyp='*', filter='')

        NOTE: Value comparison is case *insensitive* for strings.
        """
        keywords = kwd.keys()
        values = kwd.values()

        if (set(keywords).issubset(set(self.keywords))):
            # we already have the information in memory
            use_info = self.summary_info
        else:
            # we need to load information about these keywords.
            use_info = self._fits_summary(header_keywords=keywords)

        matches = np.array([True] * len(use_info))
        for key, value in zip(keywords, values):
            logger.debug('Key %s, value %s', key, value)
            logger.debug('Value in table %s', use_info[key])
            value_missing = use_info[key].mask
            logger.debug('Value missing: %s', value_missing)
            value_not_missing = np.logical_not(value_missing)
            if value == '*':
                have_this_value = value_not_missing
            elif value is not None:
                if isinstance(value, six.string_types):
                    # need to loop explicitly over array rather than using
                    # where to correctly do string comparison.
                    have_this_value = np.array([False] * len(use_info))
                    for idx, file_key_value in enumerate(use_info[key]):
                        if value_not_missing[idx]:
                            value_matches = (file_key_value.lower() ==
                                             value.lower())
                        else:
                            value_matches = False

                        have_this_value[idx] = (value_not_missing[idx] &
                                                value_matches)
                else:
                    have_this_value = value_not_missing
                    tmp = (use_info[key][value_not_missing] == value)
                    have_this_value[value_not_missing] = tmp
                    have_this_value &= value_not_missing
            else:
                # this case--when value==None--is asking for the files which
                # are missing a value for this keyword
                have_this_value = value_missing

            matches &= have_this_value

        # the numpy convention is that the mask is True for values to
        # be omitted, hence use ~matches.
        logger.debug('Matches: %s', matches)
        self.summary_info['file'].mask = ma.nomask
        self.summary_info['file'][~matches] = ma.masked

    def _fits_files_in_directory(self, extensions=None,
                                 compressed=True):
        """
        Get names of FITS files in directory, based on filename extension.

        Parameters
        ----------
        extension : list of str, optional
            List of filename extensions that are FITS files. Default is
            ``['fit', 'fits']``

        compressed : bool, optional
            If ``True``, compressed files should be included in the list
            (e.g. `.fits.gz`)

        Returns
        -------
        list
            *Names* of the files (with extension), not the full pathname.
        """
        full_extensions = extensions or ['fit', 'fits']
        if compressed:
            with_gz = [extension + '.gz' for extension in full_extensions]
            full_extensions.extend(with_gz)

        all_files = listdir(self.location)
        files = []
        for extension in full_extensions:
            files.extend(fnmatch.filter(all_files, '*' + extension))

        return files

    def _generator(self, return_type,
                   save_with_name="", save_location='',
                   clobber=False,
                   overwrite=False,
                   do_not_scale_image_data=True,
                   return_fname=False,
                   **kwd):
        """
        Generator that yields each {name} in the collection.

        If any of the parameters ``save_with_name``, ``save_location`` or
        ``overwrite`` evaluates to ``True`` the generator will write a copy of
        each FITS file it is iterating over. In other words, if
        ``save_with_name`` and/or ``save_location`` is a string with non-zero
        length, and/or ``overwrite`` is ``True``, a copy of each FITS file will
        be made.


        Parameters
        ----------
        save_with_name : str
            string added to end of file name (before extension) if
            FITS file should be saved after iteration. Unless
            `save_location` is set, files will be saved to location of
            the source files `self.location`

        save_location : str
            Directory in which to save FITS files; implies that FITS
            files will be saved. Note this provides an easy way to
            copy a directory of files--loop over the {name} with
            `save_location` set.

        overwrite : bool
            If ``True``, overwrite input FITS files.

        do_not_scale_image_data : bool
            If ``True``, prevents fits from scaling images. Default is
            ``{default_scaling}``.

        return_fname : bool, default is False
            If True, return the tuple (header, file_name) instead of just
            header.

        kwd : dict
            Any additional keywords are used to filter the items returned; see
            Examples for details.

        Returns
        -------
        {return_type}
            If ``return_fname`` is ``False``, yield the next {name} in the
            collection

        ({return_type}, str)
            If ``return_fname`` is ``True``, yield a tuple of
            ({name}, `file path`) for next the  item in the collection.

        """
        # store mask so we can reset at end--must COPY, otherwise
        # current_mask just points to the mask of summary_info
        if not self.summary_info:
            return

        current_mask = {}
        for col in self.summary_info.columns:
            current_mask[col] = self.summary_info[col].mask

        if kwd:
            self._find_keywords_by_values(**kwd)

        for full_path in self._paths():
            no_scale = do_not_scale_image_data
            hdulist = fits.open(full_path,
                                do_not_scale_image_data=no_scale)

            return_options = {'header': hdulist[0].header,
                              'hdu': hdulist[0],
                              'data': hdulist[0].data}

            try:
                yield (return_options[return_type]  # pragma: no branch
                       if (not return_fname) else
                       (return_options[return_type], full_path))
            except KeyError:
                raise ValueError('No generator for {}'.format(return_type))

            if save_location:
                destination_dir = save_location
            else:
                destination_dir = path.dirname(full_path)
            basename = path.basename(full_path)
            if save_with_name:
                base, ext = path.splitext(basename)
                basename = base + save_with_name + ext

            new_path = path.join(destination_dir, basename)

            # I really should have called the option overwrite from
            # the beginning. The hack below ensures old code works,
            # at least...
            nuke_existing = clobber or overwrite
            if (new_path != full_path) or nuke_existing:
                try:
                    hdulist.writeto(new_path, clobber=nuke_existing)
                except IOError:
                    logger.error('Error writing file %s', new_path)
                    raise
            hdulist.close()

        # reset mask
        for col in self.summary_info.columns:
            self.summary_info[col].mask = current_mask[col]

    def _paths(self):
        """
        Full path to each file.
        """
        unmasked_files = self.summary_info['file'].compressed()
        return [path.join(self.location, file_) for file_ in unmasked_files]

    def headers(self, do_not_scale_image_data=True, **kwd):
        return self._generator('header',
                               do_not_scale_image_data=do_not_scale_image_data,
                               **kwd)
    headers.__doc__ = _generator.__doc__.format(name='header',
                                                default_scaling='True',
                                                return_type='astropy.io.fits.Header')

    def hdus(self, do_not_scale_image_data=False, **kwd):
        return self._generator('hdu',
                               do_not_scale_image_data=do_not_scale_image_data,
                               **kwd)
    hdus.__doc__ = _generator.__doc__.format(name='HDU',
                                             default_scaling='False',
                                             return_type='astropy.io.fits.HDU')

    def data(self, do_not_scale_image_data=False, **kwd):
        return self._generator('data',
                               do_not_scale_image_data=do_not_scale_image_data,
                               **kwd)
    data.__doc__ = _generator.__doc__.format(name='image',
                                             default_scaling='False',
                                             return_type='numpy.ndarray')



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import logging

import numpy as np
from astropy import wcs
from astropy.io import fits

logger = logging.getLogger(__name__)


class ImageWithWCS(object):

    """
    FITS image with WCS

    A FITS images with a few convenience methods defined, including easy
    access to WCS transforms.

    Parameters
    ----------
    filepath : str
        Path to the FITS file.
    kwd : dict, optional
        All keywords are passed through to astropy.io.fits.open

    Attributes
    ----------
    data
    header
    wcs
    """

    def __init__(self, filepath, **kwd):
        self.fitsfile = fits.open(filepath, **kwd)
        self._wcs = wcs.WCS(self.header)

    @property
    def header(self):
        """FITS header for this file"""
        return self.fitsfile[0].header

    @property
    def wcs(self):
        """WCS object for this file"""
        return self._wcs

    @property
    def data(self):
        """Image data; will be scaled by default using BZERO and BSCALE"""
        return self.fitsfile[0].data

    @data.setter
    def data(self, array):
        """Set image data to numpy array"""
        self.fitsfile[0].data = array

    @property
    def shape(self):
        return self.fitsfile[0].data.shape

    def wcs_pix2world(self, pix, **kwargs):
        """
        Wrapper around astropy.wcs function that handles a single tuple
        gracefully.

        Parameters
        ----------
        pix : numpy array either of dimension 2 (e.g. [xpix, ypix]) or Nx2
            Pixel positions at which sky coordinates should be calculated.

        Returns
        -------
        numpy.ndarray with same shape as pix.
            Sky coordinates for each input pixel.
        """

        return self._wcs_wrapper(pix, self.wcs.wcs_pix2world, **kwargs)

    def wcs_world2pix(self, sky, **kwargs):
        """
        Wrapper around astropy.wcs function that handles a single tuple
        gracefully.

        Parameters
        ----------
        sky : numpy array either of dimension 2 (e.g. [xpix, ypix]) or Nx2
            Pixel positions at which sky coordinates should be calculated.

        Returns
        -------
        numpy.ndarray with same shape as sky.
            Pixel positions for each input sky coordinate.
        """
        return self._wcs_wrapper(sky, self.wcs.wcs_world2pix, **kwargs)

    def _wcs_wrapper(self, inp, transform, **kwargs):
        """"Actually handles the wrapping"""
        if not isinstance(inp, np.ndarray):
            raise TypeError('inp must be a numpy array')

        if inp.ndim == 1:
            use_pix = np.array(inp, ndmin=2)
            return_1d_array = True
        else:
            use_pix = inp
            return_1d_array = False

        ret = transform(use_pix, 1, **kwargs)  # 1 bc it is a fits file
        if return_1d_array:
            ret = ret[0]

        return ret

    def save(self, fname, overwrite=False):
        """
        Save FITS file.

        Parameters
        ----------
        fname : str
            Name of the file to save to

        overwrite : bool, optional
            Set to ``True`` to overwrite an existing file.
        """
        self.fitsfile.writeto(fname, clobber=overwrite)

    def close(self):
        """Close the file associated with this FITS image."""
        self.fitsfile.close()



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import logging

from astropy.io.fits import Header
from astropy.io.fits import PrimaryHDU
from astropy.extern import six

logger = logging.getLogger(__name__)

__all__ = ['FITSKeyword']


class FITSKeyword(object):

    """
    Represents a FITS keyword, which may have several synonyms.

    Parameters
    ----------

    name : str, optional
        Name of the keyword; case insensitive

    value : str or numeric type, optional
        Value of the keyword; this class imposes no constraints on the type of
        the keyword but if you intend to save the value in a FITS header you
        should be aware of the restrictions the FITS standard places on keyword
        values.

    comment : str, optional
        Description of the keyword.

    synonyms : str or list of str, optional
        Synonyms for this keyword. Synonyms are to look for a value in a FITS
        header and to set multiple keywords to the same value in a FITS header.
    """

    def __init__(self, name=None, value=None, comment=None, synonyms=None):
        self._hdr = Header()
        self.name = name
        self.value = value
        self.comment = comment
        if synonyms is None:
            self.synonyms = []
        else:
            self.synonyms = synonyms

    def __str__(self):
        if self.value is None:
            value_string = ''
        else:
            value_string = str(self.value)
        return ("%s = %s    / %s \n with synonyms: %s" %
                (self.name.upper(), value_string, self.comment,
                 ",".join([str(syn).upper() for syn in self.synonyms])))

    def _set_keyword_case(self, keyword):
        return keyword.upper()

    @property
    def name(self):
        """
        Primary name of the keyword.
        """
        return self._name

    @name.setter
    def name(self, keyword_name):
        self._name = self._set_keyword_case(keyword_name)
        # set synonyms again in case the new name matches one the synonyms
        self.synonyms = self.synonyms

    @property
    def synonyms(self):
        """
        List of synonyms for the keyword.
        """
        try:
            return self._synonyms
        except AttributeError:
            return None

    @synonyms.setter
    def synonyms(self, inp_synonyms):
        if not inp_synonyms:
            self._synonyms = []
            return
        if isinstance(inp_synonyms, six.string_types):
            synonym_list = [inp_synonyms]
        elif isinstance(inp_synonyms, list):
            synonym_list = set(inp_synonyms)
        else:
            raise ValueError(
                'Synonyms must either be a string or a list of strings')
        synonym_list = [self._set_keyword_case(syn) for syn in synonym_list]
        self._synonyms = [synonym for synonym in synonym_list
                          if synonym != self.name]
        return

    @property
    def names(self):
        """
        All names, including synonyms, for this keyword, as a list.
        """
        all_names = [self.name]
        if self.synonyms:
            all_names.extend(self.synonyms)
        return all_names

    def history_comment(self, with_name=None):
        """
        Produce a string describing changes to the keyword value.

        Parameters
        ----------

        with_name : str, optional
            Name to use for the keyword in the history comment. Default is the
            `name` attribute of the `Keyword`.
        """
        if with_name is None:
            with_name = self.name
        return "Updated keyword %s to value %s" % (with_name.upper(),
                                                   self.value)

    def add_to_header(self, hdu_or_header, with_synonyms=True, history=False):
        """
        Add keyword to FITS header.

        Parameters
        ----------

        hdu_or_header : astropy.io.fits.Header or astropy.io.fits.PrimaryHDU
            Header/HDU to which the keyword is to be added.

        with_synonyms : bool, optional
            Control whether a keyword is added for each of the synonyms for the
            keyword. Default is True.

        history : bool, optional
            Control whether a history comment is added to the header; if True
            a history comment is added for *each* of the keyword names added
            to the header, including synonyms.
        """
        if isinstance(hdu_or_header, PrimaryHDU):
            header = hdu_or_header.header
        elif isinstance(hdu_or_header, Header):
            header = hdu_or_header
        else:
            raise ValueError('argument must be a fits Primary HDU or header')

        header[self.name] = (self.value, self.comment)
        if history:
            header.add_history(self.history_comment())
        if with_synonyms and self.synonyms:
            for synonym in self.synonyms:
                header[synonym] = (self.value, self.comment)
                if history:
                    header.add_history(self.history_comment(with_name=synonym))

    def set_value_from_header(self, hdu_or_header):
        """
        Set value of keyword from FITS header.

        Values are obtained from the header by looking for the keyword by its
        primary name and any synonyms. If multiple values are found they are
        checked for consistency.

        Parameters
        ----------

        hdu_or_header: astropy.io.fits.Header or astrop.io.fits.PrimaryHDU
            Header from which the keyword value should be taken.

        Raises
        ------

        ValueError
            If `hdu_or_header` is of the wrong type, or the keyword
            (or synonyms) are not found in the header, or multiple
            non-identical values are found.
        """
        if isinstance(hdu_or_header, PrimaryHDU):
            header = hdu_or_header.header
        elif isinstance(hdu_or_header, Header):
            header = hdu_or_header
        else:
            raise ValueError('argument must be a fits Primary HDU or header')
        values = []
        for name in self.names:
            try:
                values.append(header[name])
            except KeyError:
                continue
        if values:
            if len(set(values)) > 1:
                error_msg = 'Found multiple values for keyword %s:\nValues: %s'
                raise ValueError(error_msg %
                                 (','.join(self.names),
                                  ','.join([str(v) for v in values])))
            self.value = values[0]
        else:
            raise ValueError('Keyword not found in header: %s' % self)



"""
Add/modify keywords in FITS files.

DESCRIPTION
-----------

    Add each of the keywords in either the ``key_file`` or specified on the
    command line to each of the files either listed in the file ``file_list`` or
    specified on the command line. If the keyword is already present its value
    is updated to the value in ``key_file``. A HISTORY comment is added to the
    header for each keyword indicating which keyword was modified.

    .. WARNING::
        This script OVERWRITES the image files in the list specified on the
        command line. There is NO WAY to override this behavior.
"""
from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

from argparse import ArgumentParser
import logging

import astropy.io.fits as fits
from astropy.table import Table
from astropy.extern import six

from . import script_helpers
from ..header_processing.fitskeyword import FITSKeyword

logger = logging.getLogger(__name__)


def add_keys(*files, **kwd):
    """
    Add keywords to a list of FITS files.

    At least one source of files must be provided and at least one source of
    keywords.

    Parameters
    ----------

    files : str, optional
        Zero or more file names, as positional arguments.

    file_list : str, optional
        Path to file which contains a list of files whose FITS keywords are to
        be modified.

    key_file : str
        Path to file which contains list of key-value pairs for FITS keywords.

    kwd : dict
        All remaining keyword arguments are interpreted as key-value pairs to
        be added as FITS keywords.

    Notes
    -----

    `file_list` should have one file per line with a header line
    at the top. A sample file list looks like this::

        File
        MyFirstFile.fits
        another_fits_file.fits
        /or/even/the/full/path/to/a/fits/file.fits

    `key_file` can be in any ASCII format easily readable by
    astropy.Table (e.g. CSV, or tab-delimited text). There needs
    to be a header line followed by, on each line, a FITS keyword
    and its value.

    All keywords in the `key_file` file are added to all of the files
    in `file_list` or the values are modified if they are already
    present in the FITS file.

    A sample `key_file` file is::

        Keyword   Value
        OBJCTDEC '+49 49 14'
        OBJCTRA '09 02 21'

    """
    file_list = kwd.pop('file_list', None)
    if file_list:
        files = Table.read(file_list, format='ascii')
        if files.colnames[0].lower() != 'file':
            raise ValueError('File list must have column named "file"')
        files.keep_columns([files.colnames[0]])
        files = [f[0] for f in files]
    key_file = kwd.pop('key_file', None)
    if key_file:
        key_table = Table.read(key_file, format='ascii')
    else:
        key_table = [(key, value) for key, value in six.iteritems(kwd)]
    for fil in files:
        logger.info('Adding keys to file %s', fil)
        fil_fits = fits.open(fil, mode='update')
        hdr = fil_fits[0].header
        for key, val in key_table:
            keyword = FITSKeyword(name=key, value=val)
            keyword.add_to_header(hdr, history=True)
            hdr[key] = val
        fil_fits.close()

#__doc__ += add_keys.__doc__


def construct_parser():
    """Add arguments to parser for this script
    """
    parser = ArgumentParser()
    script_helpers.setup_parser_help(parser, __doc__)

    file_help = 'File with list of files in which keywords are to be changed'
    parser.add_argument('--file-list', help=file_help)
    parser.add_argument('files', nargs='*',
                        help='Files in which to add/change keywords')

    key_group = parser.add_mutually_exclusive_group(required=True)

    key_group.add_argument('--key-file',
                           help='File with keywords and values to be set')

    key_group.add_argument('--key-value', nargs=2,
                           help='Keyword to add/change',)
    return parser


def main(arglist=None):
    """
    Wrapper for invoking add_keys from the command line

    Parameters
    ----------

    arglist : list of strings, optional
        If set, use this arglist instead of `sys.argv` for parsing command
        line arguments. Primarily useful for testing.
    """
    parser = construct_parser()
    args = parser.parse_args(arglist)

    if len(args.files) == 0 and args.file_list is None:
        parser.error('Must provide either files or a --file-list')
    key_dict = {}
    if args.key_value:
        key_dict[args.key_value[0]] = args.key_value[1]
    add_keys(*args.files, file_list=args.file_list, key_file=args.key_file,
             **key_dict)



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import logging
from os import path

from astropy.utils.console import _color_text

message_format = '%(message)s [%(name)s.%(funcName)s]'
console_format = '%(levelname)-18s ' + message_format
time_format = '%Y-%m-%d %H:%M:%S'
file_format = '%(asctime)s %(levelname)-8s ' + message_format

console_formatter = logging.Formatter(fmt=console_format,
                                      datefmt=time_format)


file_formatter = logging.Formatter(fmt=file_format,
                                   datefmt=time_format)


class ColorStreamHandler(logging.StreamHandler):
    def __init__(self, *args, **kwd):
        super(ColorStreamHandler, self).__init__(*args, **kwd)

    def emit(self, record):
        """
        Formatter for standard output

        Taken almost verbatim from the astropy handler logic
        """
        if record.levelno < logging.DEBUG:
            colored_name = record.levelname
        elif(record.levelno < logging.INFO):
            colored_name = _color_text(record.levelname, 'magenta')
        elif(record.levelno < logging.WARN):
            colored_name = _color_text(record.levelname, 'green')
        elif(record.levelno < logging.ERROR):
            colored_name = _color_text(record.levelname, 'blue')
        else:
            colored_name = _color_text(record.levelname, 'red')

        original_levelname = record.levelname
        record.levelname = colored_name
        message = self.format(record)
        #message = message.replace('WARNING', 'MOO')
        print (message)
        record.levelname = original_levelname
        #print (dir(record))
        #print(": " + record.message)


class FormattedFileHandler(logging.FileHandler):
    """
    Standard FileHandler with a pre-set format.
    """
    def __init__(self, *args, **kwd):
        """
        Arguments are the same as for ``FileHandler``:

            filename, mode, encoding, delay
        """
        super(FormattedFileHandler, self).__init__(*args, **kwd)
        self.setFormatter(file_formatter)


def console_handler(*args, **kwd):
    """
    Return a console handler with colored level name

    All arguments are passed to the initializer of ``ColorStreamHandler``.
    """
    console = ColorStreamHandler(*args, **kwd)
    console.setFormatter(console_formatter)
    return console


def add_file_handlers(logger, directory, script_name):
    detail_log_path = path.join(directory, script_name + '.log')
    detailed_destination_log = FormattedFileHandler(detail_log_path)
    logger.addHandler(detailed_destination_log)
    error_log_path = path.join(directory,
                               script_name + '_error.log')
    error_destination_log = FormattedFileHandler(error_log_path)
    error_destination_log.setLevel(logging.WARNING)
    logger.addHandler(error_destination_log)



from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)

INSTALL_REQUIRES = ['astropy>=1.0', 'numpy', 'ccdproc>=0.3.1']

setup(
    name='msumastro',
    version='0.8.5',
    description='Process FITS files',
    url='http://github.com/mwcraig/msumastro',
    long_description=(open('README.rst').read()),
    license='BSD 3-clause',
    author='Matt Craig',
    author_email='mcraig@mnstate.edu',
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    extras_require={
        'docs': ['numpydoc', 'sphinx-argparse', 'sphinx_rtd_theme', 'astropy-helpers'],
    },
    tests_require=['scipy', 'pytest>1.4', 'pytest-capturelog'] + INSTALL_REQUIRES,
    cmdclass={'test': PyTest},
    entry_points={
        'console_scripts': [
            ('quick_add_keys_to_file.py = '
             'msumastro.scripts.quick_add_keys_to_file:main'),
            ('run_patch.py = '
             'msumastro.scripts.run_patch:main'),
            ('run_astrometry.py = '
             'msumastro.scripts.run_astrometry:main'),
            ('run_triage.py = '
             'msumastro.scripts.run_triage:main'),
            ('run_standard_header_process.py = '
             'msumastro.scripts.run_standard_header_process:main'),
            ('sort_files.py = '
             'msumastro.scripts.sort_files:main')
        ]
    },
    classifiers=['Development Status :: 4 - Beta',
                 'License :: OSI Approved :: BSD License',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 2 :: Only',
                 'Intended Audience :: Science/Research',
                 'Topic :: Scientific/Engineering :: Astronomy'],
    )



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import logging
import subprocess
from os import path, remove, rename

from astropy.extern import six

__all__ = ['call_astrometry', 'add_astrometry']

logger = logging.getLogger(__name__)


def call_astrometry(filename, sextractor=False, feder_settings=True,
                    no_plots=True, minimal_output=True,
                    save_wcs=False, verify=None,
                    ra_dec=None, overwrite=False,
                    wcs_reference_image_center=True):
    """
    Wrapper around astrometry.net solve-field.

    Parameters
    ----------
    sextractor : bool or str, optional
        ``True`` to use `sextractor`, or a ``str`` with the
        path to sextractor.
    feder_settings : bool, optional
        Set True if you want to use plate scale appropriate for Feder
        Observatory Apogee Alta U9 camera.
    no_plots : bool, optional
        ``True`` to suppress astrometry.net generation of
        plots (pngs showing object location and more)
    minimal_output : bool, optional
        If ``True``, suppress, as separate files, output of: WCS
        header, RA/Dec object list, matching objects list, but see
        also `save_wcs`
    save_wcs : bool, optional
        If ``True``, save WCS header even if other output is suppressed
        with `minimial_output`
    verify : str, optional
        Name of a WCS header to be used as a first guess
        for the astrometry fit; if this plate solution does not work
        the solution is found as though `verify` had not been specified.
    ra_dec : list or tuple of float
        (RA, Dec); also limits search radius to 1 degree.
    overwrite : bool, optional
        If ``True``, perform astrometry even if astrometry.net files from a
        previous run are present.
    wcs_reference_image_center :
        If ``True``, force the WCS reference point in the image to be the
        image center.
    """
    solve_field = ["solve-field"]
    option_list = []

    option_list.append("--obj 40 --depth 20,40")
    if feder_settings:
        option_list.append(
            "--scale-low 0.4 --scale-high 0.56 --scale-units arcsecperpix")

    if isinstance(sextractor, six.string_types):
        option_list.append("--sextractor-path " + sextractor)
    elif sextractor:
        option_list.append("--use-sextractor")

    if no_plots:
        option_list.append("--no-plot")

    if minimal_output:
        option_list.append("--corr none --rdls none --match none")
        if not save_wcs:
            option_list.append("--wcs none")

    if ra_dec is not None:
        option_list.append("--ra %s --dec %s --radius 0.5" % ra_dec)

    if overwrite:
        option_list.append("--overwrite")

    if wcs_reference_image_center:
        option_list.append("--crpix-center")

    options = " ".join(option_list)

    solve_field.extend(options.split())

    # kludge to handle case when path of verify file contains a space--split
    # above does not work for that case.

    if verify is not None:
        solve_field.append("--verify")
        solve_field.append("%s" % verify)

    solve_field.extend([filename])
    logger.debug(' '.join(solve_field))
    try:
        solve_field_output = subprocess.check_output(solve_field,
                                                     stderr=subprocess.STDOUT)
        return_status = 0
        log_level = logging.DEBUG
    except subprocess.CalledProcessError as e:
        return_status = e.returncode
        solve_field_output = 'Output from astrometry.net:\n' + e.output
        log_level = logging.WARN
        logger.warning('Adding astrometry failed for %s', filename)
        raise e
    logger.log(log_level, solve_field_output)
    return return_status


def add_astrometry(filename, overwrite=False, ra_dec=None,
                   note_failure=False, save_wcs=False,
                   verify=None, try_builtin_source_finder=False):
    """Add WCS headers to FITS file using astrometry.net

    Parameters
    ----------
    overwrite : bool, optional
        Set ``True`` to overwrite the original file. If `False`,
        the file astrometry.net generates is kept.

    ra_dec : list or tuple of float or str
        (RA, Dec) of field center as either decimal or sexagesimal; also
        limits search radius to 1 degree.

    note_failure : bool, optional
        If ``True``, create a file with extension "failed" if astrometry.net
        fails. The "failed" file contains the error messages genreated by
        astrometry.net.

    try_builtin_source_finder : bool
        If true, try using astrometry.net's built-in source extractor if
        sextractor fails.

    save_wcs :
    verify :
        See :func:`call_astrometry`

    Returns
    -------
    bool
        ``True`` on success.

    Notes
    -----

    Tries a couple strategies before giving up: first sextractor,
    then, if that fails, astrometry.net's built-in source extractor.

    It also cleans up after astrometry.net, keeping only the new FITS
    file it generates, the .solved file, and, if desired, a ".failed" file
    for fields which it fails to solve.

    For more flexible invocation of astrometry.net, see :func:`call_astrometry`
    """
    base, ext = path.splitext(filename)

    logger.info('BEGIN ADDING ASTROMETRY on {0}'.format(filename))
    try:
        logger.debug('About to call call_astrometry')
        solved_field = (call_astrometry(filename,
                                        sextractor=True,
                                        ra_dec=ra_dec,
                                        save_wcs=save_wcs, verify=verify)
                        == 0)
    except subprocess.CalledProcessError as e:
        logger.debug('Failed with error')
        failed_details = e.output
        solved_field = False

    if (not solved_field) and try_builtin_source_finder:
        log_msg = 'Astrometry failed using sextractor, trying built-in '
        log_msg += 'source finder'
        logger.info(log_msg)
        try:
            solved_field = (call_astrometry(filename, ra_dec=ra_dec,
                                            overwrite=True,
                                            save_wcs=save_wcs, verify=verify)
                            == 0)
        except subprocess.CalledProcessError as e:
            failed_details = e.output
            solved_field = False

    if solved_field:
        logger.info('Adding astrometry succeeded')
    else:
        logger.warning('Adding astrometry failed for file %s', filename)

    if overwrite and solved_field:
        logger.info('Overwriting original file with image with astrometry')
        try:
            rename(base + '.new', filename)
        except OSError as e:
            logger.error(e)
            return False

    # whether we succeeded or failed, clean up
    try:
        remove(base + '.axy')
    except OSError:
        pass

    if solved_field:
        try:
            remove(base + '-indx.xyls')
            remove(base + '.solved')
        except OSError:
            pass

    if note_failure and not solved_field:
        try:
            f = open(base + '.failed', 'wb')
            f.write(failed_details)
            f.close()
        except IOError as e:
            logger.error('Unable to save output of astrometry.net %s', e)
            pass

    logger.info('END ADDING ASTROMETRY for %s', filename)
    return solved_field
